# This file is part of Xpra.
# Copyright (C) 2014 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# What is this workaround you ask?
# Well, pycairo can't handle raw pixel data in RGB format,
# and that's despite the documentation saying otherwise:
# http://cairographics.org/pythoncairopil/
# Because of this glaring omission, we would have to roundtrip via PNG.
# This workaround populates an image surface with RGB pixels using Cython.
# (not the most efficient implementation, but still 10 times faster than the alternative)
#
#"cairo.ImageSurface.create_for_data" is not implemented in GTK3!
# http://cairographics.org/documentation/pycairo/3/reference/surfaces.html#cairo.ImageSurface.create_for_data
# "Not yet available in Python 3"
#
#It is available in the cffi cairo bindings, which can be used instead of pycairo
# but then we can't use it from the draw callbacks:
# https://mail.gnome.org/archives/python-hackers-list/2011-December/msg00004.html
# "PyGObject just lacks the glue code that allows it to pass the statically-wrapped
# cairo.Pattern to introspected methods"


#cython: boundscheck=False

from typing import Dict
from collections.abc import Sequence

from cairo import ImageSurface

from libc.stdint cimport uintptr_t
from libc.string cimport memcpy, memset
from xpra.buffers.membuf cimport buffer_context

from xpra.util.env import envbool

RGB24_RESET_ALPHA = envbool("XPRA_RGB24_RESET_ALPHA", False)


cdef extern from "Python.h":
    ctypedef struct PyObject:
        pass
    void * PyCapsule_Import(const char *name, int no_block)

cdef extern from "cairo/cairo.h":
    ctypedef struct cairo_surface_t:
        pass

    #typedef enum _cairo_format {
    ctypedef enum cairo_format_t:
        CAIRO_FORMAT_INVALID
        CAIRO_FORMAT_ARGB32
        CAIRO_FORMAT_RGB24
        CAIRO_FORMAT_A8
        CAIRO_FORMAT_A1
        CAIRO_FORMAT_RGB16_565
        CAIRO_FORMAT_RGB30

    unsigned char * cairo_image_surface_get_data(cairo_surface_t *surface)
    cairo_format_t cairo_image_surface_get_format(cairo_surface_t *surface)

    int cairo_image_surface_get_width (cairo_surface_t *surface)
    int cairo_image_surface_get_height (cairo_surface_t *surface)
    int cairo_image_surface_get_stride (cairo_surface_t *surface)

    void cairo_surface_flush (cairo_surface_t *surface)
    void cairo_surface_mark_dirty (cairo_surface_t *surface)
    int cairo_format_stride_for_width(cairo_format_t format, int width)

cdef extern from "pycairo/py3cairo.h":
    ctypedef struct Pycairo_CAPI_t:
        pass
    ctypedef struct PycairoSurface:
        #PyObject_HEAD
        cairo_surface_t *surface
        #PyObject *base; /* base object used to create surface, or NULL */
    ctypedef PycairoSurface PycairoImageSurface


CAIRO_FORMAT: Dict[cairo_format_t, str] = {
    CAIRO_FORMAT_INVALID    : "Invalid",
    CAIRO_FORMAT_ARGB32     : "ARGB32",
    CAIRO_FORMAT_RGB24      : "RGB24",
    CAIRO_FORMAT_A8         : "A8",
    CAIRO_FORMAT_A1         : "A1",
    CAIRO_FORMAT_RGB16_565  : "RGB16_565",
    CAIRO_FORMAT_RGB30      : "RGB30",
}


cdef inline void simple_copy(uintptr_t dst, uintptr_t src, int dst_stride, int src_stride, int height) noexcept nogil:
    cdef int stride = src_stride
    cdef int pad = 0
    with nogil:
        if src_stride == dst_stride:
            memcpy(<void*> dst, <void*> src, stride*height)
        else:
            if dst_stride < src_stride:
                # cropping:
                stride = dst_stride
            else:
                # padding:
                pad = dst_stride - src_stride
            for _ in range(height):
                memcpy(<void*> dst, <void*> src, stride)
                if pad > 0:
                    memset(<void*> (dst + src_stride), 0, pad)
                src += src_stride
                dst += dst_stride


CAIRO_FORMATS: Dict[cairo_format_t, Sequence[str]] = {
    CAIRO_FORMAT_RGB24  : ("RGB", "RGBX", "BGR", "BGRX", "RGBA", "BGRA"),
    CAIRO_FORMAT_ARGB32 : ("BGRX", "BGRA", "RGBA", "RGBX"),
    CAIRO_FORMAT_RGB16_565  : ("BGR565", ),
    CAIRO_FORMAT_RGB30  : ("r210", ),
}


def make_image_surface(fmt, rgb_format: str, pixels, int width, int height, int stride) -> ImageSurface:
    if len(pixels)<height*stride:
        raise ValueError(f"pixel buffer is too small for {width}x{height} with stride={stride}:"+
                         f" only {len(pixels)} bytes, expected {height*stride}")

    cdef int x, y
    cdef int srci, dsti
    cdef const unsigned char * cbuf
    cdef cairo_surface_t * surface = NULL
    cdef unsigned char * cdata
    cdef cairo_format_t cairo_format
    cdef int istride
    cdef int iwidth
    cdef int iheight
    cdef cstride

    with buffer_context(pixels) as bc:
        # create_for_data uses the pixel data in-place, so we have to ensure that:
        # * the number of bytes per pixel is 4 (CAIRO_FORMAT_RGB24 uses 32 bits, 8 bits are unused)
        # * the pixel order is the same (whereas the slow path can do byte swapping)
        # * the rowstride matches
        if not bc.is_readonly() and rgb_format in ("BGRA", "BGRX") and fmt in (CAIRO_FORMAT_ARGB32, CAIRO_FORMAT_RGB24):
            # maybe we can just create an ImageSurface directly:
            cstride = cairo_format_stride_for_width(fmt, width)
            if stride >= cstride and (stride % 4) == 0:
                return ImageSurface.create_for_data(pixels, fmt, width, height, stride)

        image_surface = ImageSurface(fmt, width, height)
        # convert pixel_data to a C buffer:
        # convert cairo.ImageSurface python object to a cairo_surface_t
        surface = (<PycairoImageSurface *> image_surface).surface
        cairo_surface_flush(surface)
        cdata = cairo_image_surface_get_data(surface)
        #get surface attributes:
        cairo_format = cairo_image_surface_get_format(surface)
        istride    = cairo_image_surface_get_stride(surface)
        iwidth     = cairo_image_surface_get_width(surface)
        iheight    = cairo_image_surface_get_height(surface)
        if iwidth != width or iheight != height:
            raise ValueError(f"invalid image surface: expected {width}x{height} but got {iwidth}x{iheight}")
        BPP = 2 if cairo_format == CAIRO_FORMAT_RGB16_565 else 4
        if istride < iwidth * BPP:
            raise ValueError(f"invalid image stride: expected at least {iwidth*4} but got {istride}")

        cbuf = <const unsigned char *> (<uintptr_t> int(bc))
        #only deal with the formats we care about:
        if cairo_format==CAIRO_FORMAT_RGB24:
            #cairo's RGB24 format is actually stored as BGR on little endian
            if rgb_format=="BGR":
                with nogil:
                    for y in range(iheight):
                        for x in range(iwidth):
                            srci = x*3 + y*stride
                            dsti = x*4 + y*istride
                            cdata[dsti + 0] = cbuf[srci + 0]     #B
                            cdata[dsti + 1] = cbuf[srci + 1]     #G
                            cdata[dsti + 2] = cbuf[srci + 2]     #R
                            cdata[dsti + 3] = 0xff               #X
            elif rgb_format=="RGB":
                with nogil:
                    for y in range(height):
                        for x in range(width):
                            srci = x*3 + y*stride
                            dsti = x*4 + y*istride
                            cdata[dsti + 0] = cbuf[srci + 2]     #B
                            cdata[dsti + 1] = cbuf[srci + 1]     #G
                            cdata[dsti + 2] = cbuf[srci + 0]     #R
                            cdata[dsti + 3] = 0xff               #X
            elif rgb_format in ("BGRX", "BGRA"):
                if not RGB24_RESET_ALPHA:
                    with nogil:
                        simple_copy(<uintptr_t> cdata, <uintptr_t> cbuf, istride, stride, height)
                else:
                    with nogil:
                        for y in range(height):
                            for x in range(width):
                                srci = x*4 + y*stride
                                dsti = x*4 + y*istride
                                cdata[dsti + 0] = cbuf[srci + 0]     #B
                                cdata[dsti + 1] = cbuf[srci + 1]     #G
                                cdata[dsti + 2] = cbuf[srci + 2]     #R
                                cdata[dsti + 3] = 0xff               #X
            elif rgb_format in ("RGBX", "RGBA"):
                with nogil:
                    for y in range(height):
                        for x in range(width):
                            srci = x*4 + y*stride
                            dsti = x*4 + y*istride
                            cdata[dsti + 0] = cbuf[srci + 2]     #B
                            cdata[dsti + 1] = cbuf[srci + 1]     #G
                            cdata[dsti + 2] = cbuf[srci + 0]     #R
                            cdata[dsti + 3] = 0xff               #X
            else:
                raise ValueError(f"unhandled pixel format for RGB24: {rgb_format!r}")
        elif cairo_format==CAIRO_FORMAT_ARGB32:
            if rgb_format == "BGRA":
                with nogil:
                    simple_copy(<uintptr_t> cdata, <uintptr_t> cbuf, istride, stride, height)
            elif rgb_format == "RGBA":
                with nogil:
                    for y in range(height):
                        for x in range(width):
                            cdata[x*4 + 0 + y*istride] = cbuf[x*4 + 2 + y*stride]    #B
                            cdata[x*4 + 1 + y*istride] = cbuf[x*4 + 1 + y*stride]    #G
                            cdata[x*4 + 2 + y*istride] = cbuf[x*4 + 0 + y*stride]    #R
                            cdata[x*4 + 3 + y*istride] = cbuf[x*4 + 3 + y*stride]    #A
            elif rgb_format == "RGBX":
                with nogil:
                    for y in range(height):
                        for x in range(width):
                            cdata[x*4 + 0 + y*istride] = cbuf[x*4 + 2 + y*stride]    #B
                            cdata[x*4 + 1 + y*istride] = cbuf[x*4 + 1 + y*stride]    #G
                            cdata[x*4 + 2 + y*istride] = cbuf[x*4 + 0 + y*stride]    #R
                            cdata[x*4 + 3 + y*istride] = 0xff                        #A
            elif rgb_format == "BGRX":
                with nogil:
                    for y in range(height):
                        for x in range(width):
                            cdata[x*4 + 0 + y*istride] = cbuf[x*4 + 0 + y*stride]    #B
                            cdata[x*4 + 1 + y*istride] = cbuf[x*4 + 1 + y*stride]    #G
                            cdata[x*4 + 2 + y*istride] = cbuf[x*4 + 2 + y*stride]    #R
                            cdata[x*4 + 3 + y*istride] = 0xff                        #A
            elif rgb_format == "RGB":
                with nogil:
                    for y in range(height):
                        for x in range(width):
                            cdata[x*4 + 0 + y*istride] = cbuf[x*3 + 2 + y*stride]    #B
                            cdata[x*4 + 1 + y*istride] = cbuf[x*3 + 1 + y*stride]    #G
                            cdata[x*4 + 2 + y*istride] = cbuf[x*3 + 0 + y*stride]    #R
                            cdata[x*4 + 3 + y*istride] = 0xff                        #A
            elif rgb_format == "BGR":
                with nogil:
                    for y in range(height):
                        for x in range(width):
                            cdata[x*4 + 0 + y*istride] = cbuf[x*3 + 0 + y*stride]    #B
                            cdata[x*4 + 1 + y*istride] = cbuf[x*3 + 1 + y*stride]    #G
                            cdata[x*4 + 2 + y*istride] = cbuf[x*3 + 2 + y*stride]    #R
                            cdata[x*4 + 3 + y*istride] = 0xff                        #A
            else:
                raise ValueError(f"unhandled pixel format for ARGB32: {rgb_format!r}")
        elif cairo_format==CAIRO_FORMAT_RGB30:
            if rgb_format in ("r210"):
                with nogil:
                    simple_copy(<uintptr_t> cdata, <uintptr_t> cbuf, istride, stride, height)
            #UNTESTED!
            #elif rgb_format in ("BGR48"):
            #access 16-bit per pixel at a time:
            #    sbuf = <unsigned short *> cbuf
            #    #write one pixel at a time: 30-bit padded to 32
            #    idata = <unsigned long*> cdata
            #    with nogil:
            #        for y in range(height):
            #            srci = y*stride
            #            dsti = y*istride
            #            for x in range(width):
            #                idata[dsti] = (sbuf[srci] & 0x3ff) + (sbuf[srci+1] & 0x3ff)<<10 + (sbuf[srci+2] & 0x3ff)<<20
            #                srci += 3
            #                dsti += 1
            else:
                raise ValueError(f"unhandled pixel format for RGB30 {rgb_format!r}")
        elif cairo_format==CAIRO_FORMAT_RGB16_565:
            if rgb_format in ("BGR565"):
                with nogil:
                    simple_copy(<uintptr_t> cdata, <uintptr_t> cbuf, istride, stride, height)
            else:
                raise ValueError(f"unhandled pixel format for RGB16_565 {rgb_format!r}")
        else:
            raise ValueError(f"unhandled cairo format {cairo_format!r}")
    cairo_surface_mark_dirty(surface)
    return image_surface


cdef Pycairo_CAPI_t * Pycairo_CAPI
Pycairo_CAPI = <Pycairo_CAPI_t*> PyCapsule_Import("cairo.CAPI", 0)



# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from typing import Any
from weakref import WeakSet
from dataclasses import dataclass, field, asdict
from collections.abc import Callable, Iterable, Sequence

from xpra.util.objects import typedict
from xpra.util.env import envint

# noinspection PyPep8


FAST_DECODE_MIN_SPEED: int = envint("XPRA_FAST_DECODE_MIN_SPEED", 70)


# note: this is just for defining the order of encodings,
# so we have both core encodings (rgb24/rgb32) and regular encodings (rgb) in here:
PREFERRED_ENCODING_ORDER: Sequence[str] = (
    "h264", "vp9", "vp8", "mpeg4",
    "mpeg4+mp4", "h264+mp4", "vp8+webm", "vp9+webm",
    "png", "png/P", "png/L", "webp", "avif",
    "rgb", "rgb24", "rgb32", "jpeg", "jpega",
    "h265", "av1",
    "scroll",
    "grayscale",
    "stream",
)

PSEUDO_LOSSLESS_ENCODINGS: Sequence[str] = ("rgb", "png", "png/P", "png/L", "webp", "avif", "jpeg", "jpega")

TRUE_LOSSLESS_ENCODINGS: Sequence[str] = ("rgb", "png", "png/P", "png/L", "webp", "avif")

PREFERRED_REFRESH_ENCODING_ORDER: Sequence[str] = (
    "webp", "avif", "png", "png/P", "png/L", "rgb", "rgb24", "rbg32", "jpeg", "jpega",
)
STREAM_ENCODINGS: Sequence[str] = (
    "h264", "vp9", "vp8", "mpeg4",
    "mpeg4+mp4", "h264+mp4", "vp8+webm", "vp9+webm",
    "h265", "av1",
)

# encoding order for edges (usually one pixel high or wide):
EDGE_ENCODING_ORDER: Sequence[str] = (
    "rgb24", "rgb32",
    "png", "webp",
    "png/P", "png/L", "rgb", "jpeg", "jpega",
)

HELP_ORDER: Sequence[str] = (
    "auto",
    "stream",
    "grayscale",
    "h264", "h265", "av1", "vp8", "vp9", "mpeg4",
    "png", "png/P", "png/L", "webp", "avif",
    "rgb", "jpeg", "jpega",
    "scroll",
)


# value: how much smaller the output is
LOSSY_PIXEL_FORMATS: dict[str, float | int] = {
    "NV12": 2,
    "YUV420P": 2,
    "YUVA420P": 2,
    "YUV422P": 1.5,
    "YUVA422P": 1.5,
}

CSC_ALIAS: dict[str, str] = {"NV12": "YUV420P"}


def get_plane_name(pixel_format: str = "YUV420P", index: int = 0) -> str:
    return {
        "NV12": ("Y", "UV"),
    }.get(pixel_format, list(pixel_format))[index]


PIXEL_SUBSAMPLING : dict[str, Sequence[tuple[int, int]]] = {
    # NV12 is actually subsampled horizontally too - just like YUV420P
    # (but combines U and V planes so the resulting rowstride for the UV plane is the same as the Y plane):
    "NV12"      : ((1, 1), (1, 2)),
    "YUV420P"   : ((1, 1), (2, 2), (2, 2)),
    "YUVA420P"  : ((1, 1), (2, 2), (2, 2), (1, 1)),
    "YUV422P"   : ((1, 1), (2, 1), (2, 1)),
    "YUVA422P"  : ((1, 1), (2, 1), (2, 1), (1, 1)),
    "YUV444P"   : ((1, 1), (1, 1), (1, 1)),
    "YUVA444P"  : ((1, 1), (1, 1), (1, 1), (1, 1)),
    "YUV400P"   : ((1, 1), ),
    "GBRP"      : ((1, 1), (1, 1), (1, 1)),
    "GBRP9LE"   : ((1, 1), (1, 1), (1, 1)),
    "GBRP10"    : ((1, 1), (1, 1), (1, 1)),
    "YUV444P10" : ((1, 1), (1, 1), (1, 1)),
    "YUV444P16" : ((1, 1), (1, 1), (1, 1)),
}


def get_subsampling_divs(pixel_format: str) -> Sequence[tuple[int, int]]:
    # Return size dividers for the given pixel format
    #  (Y_w, Y_h), (U_w, U_h), (V_w, V_h)
    if pixel_format not in PIXEL_SUBSAMPLING:
        raise ValueError(f"invalid pixel format: {pixel_format!r}")
    return PIXEL_SUBSAMPLING[pixel_format]


def preforder(encodings: Iterable[str]) -> Sequence[str]:
    encs: set[str] = set(encodings)
    return tuple(x for x in PREFERRED_ENCODING_ORDER if x in encs)


def get_profile(options: typedict, encoding: str = "h264", csc_mode: str = "YUV420P",
                default_profile: str = "constrained-baseline") -> str:
    for x in (
        options.strget(f"{encoding}.{csc_mode}.profile"),
        options.strget(f"{encoding}.profile"),
        os.environ.get(f"XPRA_{encoding.upper()}_{csc_mode}_PROFILE"),
        os.environ.get(f"XPRA_{encoding.upper()}_PROFILE"),
        default_profile,
    ):
        if x:
            return x
    return ""


def get_x264_quality(pct: int, profile: str = "") -> int:
    if pct >= 100 and profile == "high444":
        return 0
    return 50 - (min(100, max(0, pct)) * 49 // 100)


def get_x264_preset(speed: int = 50, fast_decode: bool = False) -> int:
    if fast_decode:
        speed = max(FAST_DECODE_MIN_SPEED, speed)
    if speed > 99:
        # only allow "ultrafast" if pct > 99
        return 0
    return 5 - max(0, min(4, speed // 20))


RGB_FORMATS: Sequence[str] = (
    "XRGB",
    "BGRX",
    "ARGB",
    "BGRA",
    "RGB",
    "BGR",
    "r210",
)


class TransientCodecException(RuntimeError):
    pass


class CodecStateException(RuntimeError):
    pass


class EncodingNotSupported(RuntimeError):
    pass


@dataclass(kw_only=True)
class CodecSpec:

    codec_class     : Callable
    codec_type      : str
    input_colorspace: str = "invalid"
    output_colorspaces : Sequence[str] = ()      # ie: ("YUV420P" : "YUV420P", ...)
    quality         : int = 50
    speed           : int = 50
    size_efficiency : int = 50
    setup_cost      : int = 50
    cpu_cost        : int = 100
    gpu_cost        : int = 0
    min_w           : int = 1
    min_h           : int = 1
    max_w           : int = 4 * 1024
    max_h           : int = 4 * 1024
    can_scale       : bool = False
    score_boost     : int = 0
    width_mask      : int = 0xFFFF
    height_mask     : int = 0xFFFF
    max_instances   : int = 0
    skipped_fields : Sequence[str] = ("instances", "skipped_fields", )
    # not exported:
    instances       : WeakSet[Any] = field(default_factory=WeakSet)

    def make_instance(self) -> object:
        # pylint: disable=import-outside-toplevel
        # I can't imagine why someone would have more than this many
        # encoders or csc modules active at the same time!
        WARN_LIMIT = envint("XPRA_CODEC_INSTANCE_COUNT_WARN", 25)
        from xpra.log import Logger
        log = Logger("encoding")
        cur = self.get_instance_count()
        if 0 < self.max_instances < cur or cur >= WARN_LIMIT:
            instances = tuple(self.instances)
            log.warn(f"Warning: already {cur} active instances of {self.codec_class}:")
            try:
                import gc
                for i in instances:
                    refs = gc.get_referrers(i)
                    log.warn(f" referers({i})={refs}")
            except Exception:
                pass
        else:
            log("make_instance() %s - instance count=%s", self.codec_type, cur)
        v = self.codec_class()
        self.instances.add(v)
        return v

    def get_instance_count(self) -> int:
        return len(self.instances)

    def to_dict(self, *skip: str) -> dict[str, Any]:
        v = asdict(self)
        for k in self.skipped_fields:
            v.pop(k, None)
        for k in skip:
            v.pop(k, None)
        return v

    def get_runtime_factor(self) -> float:
        # a cost multiplier that some encoder may want to override
        # 1.0 means no change:
        mi = self.max_instances
        ic = len(self.instances)
        if ic == 0 or mi == 0:
            return 1.0                      # no problem
        if ic >= mi:
            return 0                        # not possible
        if mi > 0 and ic > 0:
            # squared slope: 50% utilisation -> value=0.75
            return max(0.0, 1.0 - (1.0*ic/mi)**2)
        return 1.0


@dataclass(kw_only=True)
class VideoSpec(CodecSpec):

    encoding           : str = "invalid"
    has_lossless_mode   : bool = False
    full_range: bool = True

    def __repr__(self):
        return f"{self.codec_type}({self.input_colorspace} to {self.encoding})"


@dataclass(kw_only=True)
class CSCSpec(CodecSpec):

    def __repr__(self):
        return f"{self.codec_type}({self.input_colorspace} to {self.output_colorspaces})"


def main():
    # pylint: disable=import-outside-toplevel
    from xpra.platform import program_context
    with program_context("Codec-Constants", "Codec Constants Info"):
        import sys
        from xpra.log import Logger, consume_verbose_argv
        consume_verbose_argv(sys.argv, "encoding")
        log = Logger("encoding")
        log.info("LOSSY_PIXEL_FORMATS=%s", LOSSY_PIXEL_FORMATS)
        log.info("PIXEL_SUBSAMPLING=%s", PIXEL_SUBSAMPLING)
        log.info("RGB_FORMATS=%s", RGB_FORMATS)


if __name__ == "__main__":
    main()
COMPRESS_FMT_PREFIX : str = "compress: %5.1fms for %4ix%-4i pixels at %4i,%-4i for wid=%-5i using %9s"
COMPRESS_RATIO      : str = " with ratio %5.1f%%  (%5iKB to %5iKB)"
COMPRESS_FMT_SUFFIX : str = ", sequence %5i, client_options=%-50s, options=%s"
COMPRESS_FMT        : str = COMPRESS_FMT_PREFIX + COMPRESS_RATIO + COMPRESS_FMT_SUFFIX
COMPRESS_FMT_DIRECT : str = (
    "compress:            %4ix%-4i pixels at %4i,%-4i for wid=%-5i using %9s" + COMPRESS_RATIO + COMPRESS_FMT_SUFFIX
)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
import socket

from xpra.log import Logger
log = Logger()

from xpra.platform.dotxpra import DotXpra
from xpra.net.bytestreams import SocketConnection
from xpra.scripts.config import make_defaults_struct


def test_DoS(client_class_constructor, args):
    """ utility method for running DoS tests
        See: test_DoS_*_client.py
    """

    assert len(args)==2, "usage: test_DoS_client :DISPLAY"
    log.enable_debug()
    opts = make_defaults_struct()
    opts.password_file = ""
    opts.encoding = "rgb24"
    opts.jpegquality = 70
    opts.quality = 70
    opts.compression_level = 1
    opts.encryption = ""
    display = sys.argv[1]
    target = DotXpra().socket_path(display)
    print("will attempt to connect to socket: %s" % target)
    sock = socket.socket(socket.AF_UNIX)
    sock.connect(target)
    conn = SocketConnection(sock, sock.getsockname(), sock.getpeername(), target, "test_DoS")
    print("socket connection=%s" % conn)
    app = client_class_constructor(conn, opts)
    try:
        app.run()
    finally:
        app.cleanup()
    print("ended")
    print("")



# This file is part of Xpra.
# Copyright (C) 2010 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any
from collections.abc import Callable, Sequence

from xpra.keyboard.common import KeyEvent
from xpra.keyboard.mask import mask_to_names, MODIFIER_MAP
from xpra.log import Logger

log = Logger("keyboard")


class KeyboardBase:

    def __init__(self):
        self.init_vars()

    def init_vars(self) -> None:
        self.modifier_keys: dict[str, str] = {}
        self.modifier_names: dict[str, str] = {}
        self.modifier_keycodes: dict[str, list[int]] = {}
        self.modifier_mappings: dict[str, Any] = {}
        # FIXME: this only allows a single modifier per mask
        # and in some cases we want to allow other modifier names
        # to use the same mask... (ie: META on OSX)
        self.modifier_map = MODIFIER_MAP

    def cleanup(self) -> None:
        self.init_vars()

    def has_bell(self) -> bool:
        return False

    def _add_modifier_mapping(self, a, b, modifier) -> None:
        # log.info("%s (%s), %s (%s)", keycode, type(keycode), keyname, type(keyname))
        if isinstance(a, int) and isinstance(b, (bytes, str)):
            self._do_add_modifier_mapping((b,), a, modifier)
        elif isinstance(a, (str, bytes)) and isinstance(b, int):
            # level = b
            self._do_add_modifier_mapping((a,), 0, modifier)
        elif isinstance(a, (bytes, str)) and isinstance(b, (bytes, str)):
            self._do_add_modifier_mapping((a, b), 0, modifier)
        elif isinstance(a, (tuple, list)) and isinstance(b, (bytes, str)):
            # ie: a=(57, 'CapsLock'), b='CapsLock'
            if len(a) == 2:
                self._add_modifier_mapping(a[0], a[1], modifier)
            self._add_modifier_mapping((b,), 0, modifier)
        elif isinstance(a, (tuple, list)) and isinstance(b, int):
            # ie: a=('CapsLock'), b=0
            self._do_add_modifier_mapping(a, 0, modifier)
        else:
            log.warn(f"Warning: unexpected key definition: {type(a)}, {type(b)}")
            log.warn(f" values: {a}, {b}")

    def _do_add_modifier_mapping(self, keynames, keycode, modifier) -> None:
        for keyname in keynames:
            self.modifier_keys[keyname] = modifier
            self.modifier_names[modifier] = keyname
            if keycode:
                keycodes = self.modifier_keycodes.setdefault(keyname, [])
                if keycode not in keycodes:
                    keycodes.append(keycode)

    def set_modifier_mappings(self, mappings: dict[str, Any]) -> None:
        log("set_modifier_mappings({mappings})")
        self.modifier_mappings = mappings
        self.modifier_keys = {}
        self.modifier_names = {}
        self.modifier_keycodes = {}
        for modifier, keys in mappings.items():
            for a, b in keys:
                self._add_modifier_mapping(a, b, modifier)
        log(f"modifier_keys={self.modifier_keys}")
        log(f"modifier_names={self.modifier_names}")
        log(f"modifier_keycodes={self.modifier_keycodes}")

    def mask_to_names(self, mask) -> list[str]:
        # mask can be an int, or a `ModifierType` enum:
        return mask_to_names(int(mask), self.modifier_map)

    def get_keymap_modifiers(self) -> tuple[dict, list[str], list[str]]:
        """
            ask the server to manage capslock ('lock') which can be missing from mouse events
            (or maybe this is virtualbox causing it?)
        """
        return {}, [], ["lock"]

    def get_keymap_spec(self) -> dict[str, Any]:
        return {}

    def get_x11_keymap(self) -> dict[int, list[str]]:
        return {}

    def get_layout_spec(self) -> tuple[str, str, Sequence[str], str, Sequence[str], str]:
        return "", "", (), "", (), ""

    def get_keyboard_repeat(self) -> tuple[int, int] | None:
        return None

    def update_modifier_map(self, mod_meanings) -> None:
        log(f"update_modifier_map({mod_meanings})")
        self.modifier_map = MODIFIER_MAP

    def process_key_event(self, send_key_action_cb: Callable, wid: int, key_event: KeyEvent):
        # default is to just send it as-is:
        send_key_action_cb(wid, key_event)



# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

#cython: boundscheck=False, wraparound=False, initializedcheck=False, cdivision=True, always_allow_keywords=False

from typing import List, Tuple
from xpra.common import SizedBuffer

from xpra.util.env import first_time
from xpra.buffers.membuf cimport getbuf, MemBuf, buffer_context  # pylint: disable=syntax-error

from libc.stdint cimport uintptr_t, uint32_t, uint16_t, uint8_t

import struct
from xpra.log import Logger
log = Logger("encoding")


cdef inline unsigned int round8up(unsigned int n) nogil:
    return (n + 7) & ~7


cdef inline unsigned char clamp(int v) nogil:
    if v>255:
        return 255
    return <unsigned char> v


emptymem = memoryview(b"")


def bgr565_to_rgbx(buf: SizedBuffer) -> memoryview:
    assert len(buf) % 2 == 0, "invalid buffer size: %s is not a multiple of 2" % len(buf)
    cdef const uint16_t *rgb565
    with buffer_context(buf) as bc:
        rgb565 = <const uint16_t*> (<uintptr_t> int(bc))
        return bgr565data_to_rgbx(rgb565, len(bc))


cdef memoryview bgr565data_to_rgbx(const uint16_t* rgb565, const int rgb565_len):
    if rgb565_len <= 0:
        return emptymem
    assert rgb565_len>0 and rgb565_len % 2 == 0, "invalid buffer size: %s is not a multiple of 2" % rgb565_len
    cdef MemBuf output_buf = getbuf(rgb565_len*2, 0)
    cdef uint32_t *rgbx = <uint32_t*> output_buf.get_mem()
    cdef uint16_t v
    cdef unsigned int l = rgb565_len//2
    with nogil:
        for i in range(l):
            v = rgb565[i]
            rgbx[i] = (<uint32_t> 0xff000000) | (((v & 0xF800) >> 8) | ((v & 0x07E0) << 5) | ((v & 0x001F) << 19))
    return memoryview(output_buf)


def bgr565_to_rgb(buf: SizedBuffer) -> memoryview:
    assert len(buf) % 2 == 0, "invalid buffer size: %s is not a multiple of 2" % len(buf)
    cdef const uint16_t* rgb565
    with buffer_context(buf) as bc:
        rgb565 = <const uint16_t*> (<uintptr_t> int(bc))
        return bgr565data_to_rgb(rgb565, len(bc))


cdef memoryview bgr565data_to_rgb(const uint16_t* rgb565, const int rgb565_len):
    if rgb565_len <= 0:
        return emptymem
    assert rgb565_len>0 and rgb565_len % 2 == 0, "invalid buffer size: %s is not a multiple of 2" % rgb565_len
    cdef MemBuf output_buf = getbuf(rgb565_len*3//2, 0)
    cdef uint8_t *rgb = <uint8_t*> output_buf.get_mem()
    cdef uint32_t v, i
    cdef unsigned int l = rgb565_len//2
    with nogil:
        for i in range(l):
            v = rgb565[i]
            rgb[0] = (v & 0xF800) >> 8
            rgb[1] = (v & 0x07E0) >> 3
            rgb[2] = (v & 0x001F) << 3
            rgb += 3
    return memoryview(output_buf)


def r210_to_rgba(buf: SizedBuffer,
                 const unsigned int w, const unsigned int h,
                 const unsigned int src_stride, const unsigned int dst_stride) -> memoryview:
    assert w*4<=src_stride, "invalid source stride %i for width %i" % (src_stride, w)
    assert w*4<=dst_stride, "invalid destination stride %i for width %i" % (dst_stride, w)
    cdef unsigned int* r210
    with buffer_context(buf) as bc:
        assert len(bc)>=<Py_ssize_t>(h*src_stride), "source buffer is %i bytes, which is too small for %ix%i" % (len(bc), src_stride, h)
        r210 = <unsigned int*> (<uintptr_t> int(bc))
        return r210data_to_rgba(r210, w, h, src_stride, dst_stride)


cdef memoryview r210data_to_rgba(unsigned int* r210,
                      const unsigned int w, const unsigned int h,
                      const unsigned int src_stride, const unsigned int dst_stride):
    cdef MemBuf output_buf = getbuf(h*dst_stride, 0)
    cdef unsigned int* rgba = <unsigned int*> output_buf.get_mem()
    cdef unsigned int v, x, y = 0
    with nogil:
        for y in range(h):
            for x in range(w):
                v = r210[x]
                rgba[x] = (v&0x3fc00000) >> 22 | (v&0x000ff000) >> 4 | (v&0x000003fc) << 14 | ((v>>30)*85)<<24
            r210 = <unsigned int*> ((<uintptr_t> r210) + src_stride)
            rgba = <unsigned int*> ((<uintptr_t> rgba) + dst_stride)
    return memoryview(output_buf)


def r210_to_rgbx(buf: SizedBuffer,
                 const unsigned int w, const unsigned int h,
                 const unsigned int src_stride, const unsigned int dst_stride) -> memoryview:
    assert buf, "no buffer"
    assert w*4<=src_stride, "invalid source stride %i for width %i" % (src_stride, w)
    assert w*4<=dst_stride, "invalid destination stride %i for width %i" % (dst_stride, w)
    cdef unsigned int* r210
    with buffer_context(buf) as bc:
        assert len(bc)>=<Py_ssize_t>(h*src_stride), "source buffer is %i bytes, which is too small for %ix%i" % (len(bc), src_stride, h)
        r210 = <unsigned int*> (<uintptr_t> int(bc))
        return r210data_to_rgbx(r210, w, h, src_stride, dst_stride)


cdef memoryview r210data_to_rgbx(unsigned int* r210,
                      const unsigned int w, const unsigned int h,
                      const unsigned int src_stride, const unsigned int dst_stride):
    cdef MemBuf output_buf = getbuf(h*dst_stride, 0)
    cdef unsigned int* rgbx = <unsigned int*> output_buf.get_mem()
    cdef unsigned int v, x, y = 0
    with nogil:
        for y in range(h):
            for x in range(w):
                v = r210[x]
                rgbx[x] = (v&0x3fc00000) >> 22 | (v&0x000ff000) >> 4 | (v&0x000003fc) << 14 | <unsigned int> 0xff000000
            r210 = <unsigned int*> ((<uintptr_t> r210) + src_stride)
            rgbx = <unsigned int*> ((<uintptr_t> rgbx) + dst_stride)
    return memoryview(output_buf)


def r210_to_rgb(buf: SizedBuffer,
                const unsigned int w, const unsigned int h,
                const unsigned int src_stride, const unsigned int dst_stride) -> memoryview:
    assert buf, "no buffer"
    assert w*4<=src_stride, "invalid source stride %i for width %i" % (src_stride, w)
    assert w*3<=dst_stride, "invalid destination stride %i for width %i" % (dst_stride, w)
    cdef unsigned int* r210
    with buffer_context(buf) as bc:
        assert len(bc)>=<Py_ssize_t>(h*src_stride), "source buffer is %i bytes, which is too small for %ix%i" % (len(bc), src_stride, h)
        r210 = <unsigned int*> (<uintptr_t> int(bc))
        return r210data_to_rgb(r210, w, h, src_stride, dst_stride)


# white:  3fffffff
# red:    3ff00000
# green:     ffc00
# blue:        3ff
# black:         0
cdef memoryview r210data_to_rgb(unsigned int* r210,
                     const unsigned int w, const unsigned int h,
                     const unsigned int src_stride, const unsigned int dst_stride):
    cdef MemBuf output_buf = getbuf(h*dst_stride, 0)
    cdef unsigned char* rgba = <unsigned char*> output_buf.get_mem()
    cdef unsigned int i, v, y = 0
    with nogil:
        for y in range(h):
            i = y*dst_stride
            for x in range(w):
                v = r210[x]
                rgba[i+2] = (v&0x000003ff) >> 2
                rgba[i+1] = (v&0x000ffc00) >> 12
                rgba[i]   = (v&0x3ff00000) >> 22
                i += 3
            r210 = <unsigned int*> ((<uintptr_t> r210) + src_stride)
    return memoryview(output_buf)


def bgrx_to_rgb(buf: SizedBuffer) -> memoryview:
    assert len(buf) % 4 == 0, "invalid buffer size: %s is not a multiple of 4" % len(buf)
    cdef const unsigned int* bgrx
    with buffer_context(buf) as bc:
        bgrx = <const unsigned int*> (<uintptr_t> int(bc))
        return bgrxdata_to_rgb(bgrx, len(bc))


cdef memoryview bgrxdata_to_rgb(const unsigned int *bgrx, const int bgrx_len):
    if bgrx_len <= 0:
        return None
    assert bgrx_len>0 and bgrx_len % 4 == 0, "invalid buffer size: %s is not a multiple of 4" % bgrx_len
    #number of pixels:
    cdef int mi = bgrx_len//4
    #3 bytes per pixel:
    cdef MemBuf output_buf = getbuf(mi*3, 0)
    cdef unsigned char* rgb = <unsigned char*> output_buf.get_mem()
    cdef int si = 0, di = 0
    cdef unsigned int p
    with nogil:
        for si in range(mi):
            p = bgrx[si]
            rgb[di]   = p & 0xFF                #R
            rgb[di+1] = (p>>8) & 0xFF           #G
            rgb[di+2] = (p>>16) & 0xFF          #B
            di += 3
    return memoryview(output_buf)


def rgb_to_bgrx(buf: SizedBuffer) -> memoryview:
    assert len(buf) % 3 == 0, "invalid buffer size: %s is not a multiple of 3" % len(buf)
    cdef const unsigned char* rgb
    with buffer_context(buf) as bc:
        rgb = <const unsigned char*> (<uintptr_t> int(bc))
        return rgbdata_to_bgrx(rgb, len(bc))


cdef memoryview rgbdata_to_bgrx(const unsigned char *rgb, const int rgb_len):
    if rgb_len <= 0:
        return None
    assert rgb_len>0 and rgb_len % 3 == 0, "invalid buffer size: %s is not a multiple of 3" % rgb_len
    #number of pixels:
    cdef int mi = rgb_len//3
    #3 bytes per pixel:
    cdef MemBuf output_buf = getbuf(mi*4, 0)
    cdef unsigned int* bgrx = <unsigned int*> output_buf.get_mem()
    cdef int si = 0, di = 0
    cdef unsigned int p
    with nogil:
        while di < mi:
            bgrx[di] = rgb[si] + (rgb[si+1]<<8) + (rgb[si+2]<<16)
            di += 1
            si += 3
    return memoryview(output_buf)


def bgrx_to_l(buf: SizedBuffer) -> memoryview:
    assert len(buf) % 4 == 0, "invalid buffer size: %s is not a multiple of 4" % len(buf)
    cdef const unsigned int* bgrx
    with buffer_context(buf) as bc:
        bgrx = <const unsigned int*> (<uintptr_t> int(bc))
        return bgrxdata_to_l(bgrx, len(bc))


cdef memoryview bgrxdata_to_l(const unsigned int *bgrx, const int bgrx_len):
    if bgrx_len <= 0:
        return emptymem
    assert bgrx_len>0 and bgrx_len % 4 == 0, "invalid buffer size: %s is not a multiple of 4" % bgrx_len
    #number of pixels:
    cdef int mi = bgrx_len//4
    #3 bytes per pixel:
    cdef MemBuf output_buf = getbuf(mi, 0)
    cdef unsigned char* l = <unsigned char*> output_buf.get_mem()
    cdef int i = 0
    cdef unsigned int p
    cdef unsigned char r, g, b
    with nogil:
        for i in range(mi):
            p = bgrx[i]
            r = p & 0xFF                #R
            g = (p>>8) & 0xFF           #G
            b = (p>>16) & 0xFF          #B
            l[i] = (r*3+b+g*4)>>3
    return memoryview(output_buf)


def bgr_to_l(buf: SizedBuffer) -> memoryview:
    assert len(buf) % 3 == 0, "invalid buffer size: %s is not a multiple of 3" % len(buf)
    cdef const unsigned char* bgr
    with buffer_context(buf) as bc:
        bgr = <const unsigned char*> (<uintptr_t> int(bc))
        return rgbdata_to_l(bgr, len(bc), 2, 1, 0)


def rgb_to_l(buf: SizedBuffer) -> memoryview:
    assert len(buf) % 3 == 0, "invalid buffer size: %s is not a multiple of 3" % len(buf)
    cdef const unsigned char* rgb
    with buffer_context(buf) as bc:
        rgb = <const unsigned char*> (<uintptr_t> int(bc))
        return rgbdata_to_l(rgb, len(bc), 0, 1, 2)


cdef memoryview rgbdata_to_l(const unsigned char *rgb, const int rgb_len,
                  const unsigned char rindex, const unsigned char gindex, const unsigned char bindex):
    if rgb_len <= 0:
        return None
    assert rgb_len>0 and rgb_len % 3 == 0, "invalid buffer size: %s is not a multiple of 3" % rgb_len
    #number of pixels:
    cdef int mi = rgb_len//3
    #3 bytes per pixel:
    cdef MemBuf output_buf = getbuf(mi, 0)
    cdef unsigned char* l = <unsigned char*> output_buf.get_mem()
    cdef int i = 0
    cdef unsigned char r, g, b
    with nogil:
        while i < mi:
            r = rgb[i+rindex]
            g = rgb[i+gindex]
            b = rgb[i+bindex]
            l[i] = (r*3+b+g*4)>>3
            i += 3
    return memoryview(output_buf)


def bgra_to_la(buf: SizedBuffer) -> memoryview:
    assert len(buf) % 4 == 0, "invalid buffer size: %s is not a multiple of 4" % len(buf)
    cdef const unsigned int* bgra
    with buffer_context(buf) as bc:
        bgra = <const unsigned int*> (<uintptr_t> int(bc))
        return bgradata_to_la(bgra, len(bc))


cdef memoryview bgradata_to_la(const unsigned int *bgra, const int bgra_len):
    if bgra_len <= 0:
        return emptymem
    assert bgra_len>0 and bgra_len % 4 == 0, "invalid buffer size: %s is not a multiple of 4" % bgra_len
    #number of pixels:
    cdef int mi = bgra_len//4
    #3 bytes per pixel:
    cdef MemBuf output_buf = getbuf(mi*2, 0)
    cdef unsigned char* la = <unsigned char*> output_buf.get_mem()
    cdef int si = 0, di = 0
    cdef unsigned int p
    cdef unsigned char r, g, b, a
    with nogil:
        for si in range(mi):
            p = bgra[si]
            r = p & 0xFF
            g = (p>>8) & 0xFF
            b = (p>>16) & 0xFF
            a = (p>>24) & 0xFF
            la[di] = (r*3+b+g*4)>>3
            la[di+1] = a
            di += 2
    return memoryview(output_buf)


def argb_to_rgba(buf: SizedBuffer) -> memoryview:
    assert len(buf) % 4 == 0, "invalid buffer size: %s is not a multiple of 4" % len(buf)
    cdef const unsigned int* argb
    with buffer_context(buf) as bc:
        argb = <const unsigned int*> (<uintptr_t> int(bc))
        return argbdata_to_rgba(argb, len(bc))


cdef memoryview argbdata_to_rgba(const unsigned int* argb, const int argb_len):
    if argb_len <= 0:
        return emptymem
    assert argb_len>0 and argb_len % 4 == 0, "invalid buffer size: %s is not a multiple of 4" % argb_len
    cdef int mi = argb_len//4
    cdef MemBuf output_buf = getbuf(argb_len, 0)
    cdef unsigned int* rgba = <unsigned int*> output_buf.get_mem()
    cdef int i = 0
    cdef unsigned int p
    with nogil:
        for i in range(mi):
            p = argb[i]
            rgba[i] = p>>8 | (p&0xff)<<24
    return memoryview(output_buf)


def argb_to_rgb(buf: SizedBuffer) -> memoryview:
    assert len(buf) % 4 == 0, "invalid buffer size: %s is not a multiple of 4" % len(buf)
    cdef const unsigned int* argb
    with buffer_context(buf) as bc:
        argb = <const unsigned int*> (<uintptr_t> int(bc))
        return argbdata_to_rgb(argb, len(bc))


cdef memoryview argbdata_to_rgb(const unsigned int* argb, const int argb_len):
    if argb_len <= 0:
        return emptymem
    assert argb_len>0 and argb_len % 4 == 0, "invalid buffer size: %s is not a multiple of 4" % argb_len
    #number of pixels:
    cdef int mi = argb_len//4
    #3 bytes per pixel:
    cdef MemBuf output_buf = getbuf(mi*3, 0)
    cdef unsigned char* rgb = <unsigned char*> output_buf.get_mem()
    cdef int si = 0, di = 0
    cdef unsigned int p
    with nogil:
        for si in range(mi):
            p = argb[si]
            rgb[di]   = (p>>8)&0xFF             #R
            rgb[di+1] = (p>>16)&0xFF            #G
            rgb[di+2] = (p>>24)&0xFF            #B
            di += 3
    return memoryview(output_buf)


def bgra_to_rgb222(buf: SizedBuffer) -> memoryview:
    assert len(buf) % 4 == 0, "invalid buffer size: %s is not a multiple of 4" % len(buf)
    cdef const unsigned char* bgra
    with buffer_context(buf) as bc:
        bgra = <const unsigned char*> (<uintptr_t> int(bc))
        return bgradata_to_rgb222(bgra, len(bc))


cdef memoryview bgradata_to_rgb222(const unsigned char* bgra, const int bgra_len):
    if bgra_len <= 0:
        return emptymem
    assert bgra_len>0 and bgra_len % 4 == 0, "invalid buffer size: %s is not a multiple of 4" % bgra_len
    #number of pixels:
    cdef int mi = bgra_len//4                #@DuplicateSignature
    #1 byte per pixel:
    cdef MemBuf output_buf = getbuf(mi, 0)
    cdef unsigned char* rgb = <unsigned char*> output_buf.get_mem()
    cdef int di = 0, si = 0                  #@DuplicateSignature
    with nogil:
        while si < bgra_len:
            rgb[di] = ((bgra[si+2]>>2) & 0x30) | ((bgra[si+1]>>4) & 0xC) | ((bgra[si]>>6) & 0x3)
            di += 1
            si += 4
    return memoryview(output_buf)


def bgra_to_rgb(buf: SizedBuffer) -> memoryview:
    assert len(buf) % 4 == 0, "invalid buffer size: %s is not a multiple of 4" % len(buf)
    cdef const unsigned int* bgra
    with buffer_context(buf) as bc:
        bgra = <const unsigned int*> (<uintptr_t> int(bc))
        return bgradata_to_rgb(bgra, len(bc))


cdef memoryview bgradata_to_rgb(const unsigned int* bgra, const int bgra_len):
    if bgra_len <= 0:
        return emptymem
    assert bgra_len>0 and bgra_len % 4 == 0, "invalid buffer size: %s is not a multiple of 4" % bgra_len
    #number of pixels:
    cdef int mi = bgra_len//4
    #3 bytes per pixel:
    cdef MemBuf output_buf = getbuf(mi*3, 0)
    cdef unsigned char* rgb = <unsigned char*> output_buf.get_mem()
    cdef int di = 0, si = 0
    cdef unsigned int p
    with nogil:
        for si in range(mi):
            p = bgra[si]
            rgb[di]   = (p>>16) & 0xFF          #R
            rgb[di+1] = (p>>8) & 0xFF           #G
            rgb[di+2] = p & 0xFF                #B
            di += 3
    return memoryview(output_buf)


def bgra_to_rgba(buf: SizedBuffer) -> memoryview:
    assert len(buf) % 4 == 0, "invalid buffer size: %s is not a multiple of 4" % len(buf)
    cdef const unsigned int* bgra
    with buffer_context(buf) as bc:
        bgra = <const unsigned int*> (<uintptr_t> int(bc))
        return bgradata_to_rgba(bgra, len(bc))


cdef memoryview bgradata_to_rgba(const unsigned int* bgra, const int bgra_len):
    if bgra_len <= 0:
        return emptymem
    assert bgra_len>0 and bgra_len % 4 == 0, "invalid buffer size: %s is not a multiple of 4" % bgra_len
    cdef int mi = bgra_len//4
    cdef MemBuf output_buf = getbuf(bgra_len, 0)
    cdef unsigned int* rgba = <unsigned int*> output_buf.get_mem()
    cdef int i = 0
    cdef unsigned int p
    with nogil:
        for i in range(mi):
            p = bgra[i]
            rgba[i] = (p>>16) & 0xff | p & 0xff00 | (p & 0xff)<<16 | p&(<unsigned int>0xff000000)
    return memoryview(output_buf)


def rgba_to_bgra(buf: SizedBuffer) -> memoryview:
    #same: just a swap
    return bgra_to_rgba(buf)


def bgra_to_rgbx(buf: SizedBuffer) -> memoryview:
    assert len(buf) % 4 == 0, "invalid buffer size: %s is not a multiple of 4" % len(buf)
    cdef const unsigned char* bgra
    with buffer_context(buf) as bc:
        bgra = <const unsigned char*> (<uintptr_t> int(bc))
        return bgradata_to_rgbx(bgra, len(bc))


cdef memoryview bgradata_to_rgbx(const unsigned char* bgra, const int bgra_len):
    if bgra_len <= 0:
        return emptymem
    assert bgra_len>0 and bgra_len % 4 == 0, "invalid buffer size: %s is not a multiple of 4" % bgra_len
    #same number of bytes:
    cdef MemBuf output_buf = getbuf(bgra_len, 0)
    cdef unsigned char* rgbx = <unsigned char*> output_buf.get_mem()
    cdef int i = 0                      #@DuplicateSignature
    with nogil:
        while i < bgra_len:
            rgbx[i]   = bgra[i+2]           #R
            rgbx[i+1] = bgra[i+1]           #G
            rgbx[i+2] = bgra[i]             #B
            rgbx[i+3] = 0xff                #X
            i += 4
    return memoryview(output_buf)


def premultiply_argb(buf: SizedBuffer) -> memoryview:
    assert len(buf) % 4 == 0, "invalid buffer size: %s is not a multiple of 4" % len(buf)
    # b is a Python buffer object
    cdef unsigned int *argb
    with buffer_context(buf) as bc:
        argb = <unsigned int*> (<uintptr_t> int(bc))
        return do_premultiply_argb(argb, len(bc))


cdef memoryview do_premultiply_argb(unsigned int *buf, Py_ssize_t argb_len):
    # cbuf contains non-premultiplied ARGB32 data in native-endian.
    # We convert to premultiplied ARGB32 data
    cdef unsigned char a, r, g, b                #@DuplicateSignature
    cdef unsigned int argb                      #@DuplicateSignature
    assert argb_len>0 and argb_len % 4 == 0, "invalid buffer size: %s is not a multiple of 4" % argb_len
    cdef MemBuf output_buf = getbuf(argb_len, 0)
    cdef unsigned int* argb_out = <unsigned int*> output_buf.get_mem()
    cdef int i                                  #@DuplicateSignature
    with nogil:
        for 0 <= i < argb_len / 4:
            argb = buf[i]
            a = (argb >> 24) & 0xff
            r = (argb >> 16) & 0xff
            r = r * a // 255
            g = (argb >> 8) & 0xff
            g = g * a // 255
            b = (argb >> 0) & 0xff
            b = b * a // 255
            argb_out[i] = (a << 24) | (r << 16) | (g << 8) | (b << 0)
    return memoryview(output_buf)


def unpremultiply_argb(buf) -> memoryview:
    assert len(buf) % 4 == 0, "invalid buffer size: %s is not a multiple of 4" % len(buf)
    cdef unsigned int *argb
    with buffer_context(buf) as bc:
        argb = <unsigned int*> (<uintptr_t> int(bc))
        return do_unpremultiply_argb(argb, len(bc))


# precalculate indexes in native endianness:
tmp = struct.pack(b"=L", 0 + 1*(2**8) + 2*(2**16) + 3*(2**24))
# little endian will give 0, 1, 2, 3
# big endian should give 3, 2, 1, 0 (untested)
cdef unsigned char B = tmp.index(b'\0')
cdef unsigned char G = tmp.index(b'\1')
cdef unsigned char R = tmp.index(b'\2')
cdef unsigned char A = tmp.index(b'\3')


cdef memoryview do_unpremultiply_argb(unsigned int * argb_in, Py_ssize_t argb_len):
    # cbuf contains non-premultiplied ARGB32 data in native-endian.
    # We convert to premultiplied ARGB32 data
    cdef unsigned char a, r, g, b                #@DuplicateSignature
    cdef unsigned int argb                      #@DuplicateSignature
    assert argb_len>0 and argb_len % 4 == 0, "invalid buffer size: %s is not a multiple of 4" % argb_len
    cdef MemBuf output_buf = getbuf(argb_len, 0)
    cdef unsigned char* argb_out = <unsigned char*> output_buf.get_mem()
    cdef int i                                  #@DuplicateSignature
    with nogil:
        for 0 <= i < argb_len // 4:
            argb = argb_in[i]
            a = (argb >> 24) & 0xff
            r = (argb >> 16) & 0xff
            g = (argb >> 8) & 0xff
            b = (argb >> 0) & 0xff
            if a!=0:
                r = clamp(r * 255 // a)
                g = clamp(g * 255 // a)
                b = clamp(b * 255 // a)
            else:
                r = 0
                g = 0
                b = 0
            #we could use struct pack to avoid endianness issues
            #but this is python 2.5 onwards only and is probably slower:
            #struct.pack_into(b"=BBBB", argb_out, i*4, b, g, r, a)
            argb_out[i*4+B] = b
            argb_out[i*4+G] = g
            argb_out[i*4+R] = r
            argb_out[i*4+A] = a
    return memoryview(output_buf)


def alpha(image) -> memoryview:
    pixel_format = image.get_pixel_format()
    cdef char i = pixel_format.find("A")
    if i<0 or i>=4:
        return emptymem
    pixels = image.get_pixels()
    assert pixels, "failed to get pixels from %s" % image
    assert len(pixels) % 4 == 0, "invalid buffer size: %s is not a multiple of 4" % len(pixels)
    cdef const unsigned char* rgba
    with buffer_context(pixels) as bc:
        rgba = <const unsigned char*> (<uintptr_t> int(bc))
        return alpha_data(rgba, len(bc), i)


cdef memoryview alpha_data(const unsigned char* rgba, const int rgba_len, const char index):
    if rgba_len <= 0:
        return None
    assert rgba_len>0 and rgba_len % 4 == 0, "invalid buffer size: %s is not a multiple of 4" % rgba_len
    cdef MemBuf output_buf = getbuf(rgba_len//4, 0)
    cdef unsigned char* alpha = <unsigned char*> output_buf.get_mem()
    cdef int di = 0, si = index
    cdef unsigned int p
    with nogil:
        while si < rgba_len:
            alpha[di] = rgba[si]
            si += 4
            di += 1
    return memoryview(output_buf)


def argb_swap(image, rgb_formats, supports_transparency=False) -> bool:
    """ use the argb codec to do the RGB byte swapping """
    pixel_format = image.get_pixel_format()
    #try to fallback to argb module
    #if we have one of the target pixel formats:
    pixels = image.get_pixels()
    assert pixels, "failed to get pixels from %s" % image
    cdef unsigned int rs = image.get_rowstride()
    cdef unsigned int w
    cdef unsigned int h
    def checkstride(div=4):
        if rs%div:
            raise ValueError(f"invalid rowstride for {pixel_format}, {rs} is not a multiple of {div}")
    if pixel_format=="r210":
        checkstride()
        #r210 never contains any transparency at present
        #if supports_transparency and "RGBA" in rgb_formats:
        #    log("argb_swap: r210_to_rgba for %s on %s", pixel_format, type(pixels))
        #    image.set_pixels(r210_to_rgba(pixels))
        #    image.set_pixel_format("RGBA")
        #    return True
        w = image.get_width()
        h = image.get_height()
        if "RGB" in rgb_formats:
            log("argb_swap: r210_to_rgb for %s on %s", pixel_format, type(pixels))
            image.set_pixels(r210_to_rgb(pixels, w, h, rs, w*3))
            image.set_pixel_format("RGB")
            image.set_rowstride(w*3)
            return True
        if "RGBX" in rgb_formats:
            log("argb_swap: r210_to_rgbx for %s on %s", pixel_format, type(pixels))
            image.set_pixels(r210_to_rgbx(pixels, w, h, rs, w*4))
            image.set_pixel_format("RGBX")
            image.set_rowstride(w*4)
            return True
    elif pixel_format=="BGR565":
        checkstride(2)
        if "RGB" in rgb_formats:
            log("argb_swap: bgr565_to_rgb for %s on %s", pixel_format, type(pixels))
            image.set_pixels(bgr565_to_rgb(pixels))
            image.set_pixel_format("RGB")
            image.set_rowstride(rs*3//2)
            return True
        if "RGBX" in rgb_formats:
            log("argb_swap: bgr565_to_rgbx for %s on %s", pixel_format, type(pixels))
            image.set_pixels(bgr565_to_rgbx(pixels))
            image.set_pixel_format("RGBX")
            image.set_rowstride(rs*2)
            return True
    elif pixel_format in ("BGRX", "BGRA"):
        checkstride()
        if pixel_format=="BGRX" and "L" in rgb_formats:
            log("argb_swap: bgrx_to_l for %s on %s", pixel_format, type(pixels))
            image.set_pixels(bgrx_to_l(pixels))
            image.set_pixel_format("L")
            return True
        if pixel_format=="BGRA" and "LA" in rgb_formats:
            log("argb_swap: bgra_to_la for %s on %s", pixel_format, type(pixels))
            image.set_pixels(bgra_to_la(pixels))
            image.set_pixel_format("LA")
            return True
        if pixel_format=="BGRA" and supports_transparency and "RGBA" in rgb_formats:
            log("argb_swap: bgra_to_rgba for %s on %s", pixel_format, type(pixels))
            image.set_pixels(bgra_to_rgba(pixels))
            image.set_pixel_format("RGBA")
            return True
        if "RGB" in rgb_formats:
            log("argb_swap: bgra_to_rgb for %s on %s", pixel_format, type(pixels))
            image.set_pixels(bgra_to_rgb(pixels))
            image.set_pixel_format("RGB")
            image.set_rowstride(rs*3//4)
            return True
        if "RGBX" in rgb_formats:
            log("argb_swap: bgra_to_rgbx for %s on %s", pixel_format, type(pixels))
            image.set_pixels(bgra_to_rgbx(pixels))
            image.set_pixel_format("RGBX")
            return True
    elif pixel_format in ("XRGB", "ARGB"):
        checkstride()
        if pixel_format=="ARGB" and supports_transparency and "RGBA" in rgb_formats:
            log("argb_swap: argb_to_rgba for %s on %s", pixel_format, type(pixels))
            image.set_pixels(argb_to_rgba(pixels))
            image.set_pixel_format("RGBA")
            return True
        if "RGB" in rgb_formats:
            log("argb_swap: argb_to_rgb for %s on %s", pixel_format, type(pixels))
            image.set_pixels(argb_to_rgb(pixels))
            image.set_pixel_format("RGB")
            image.set_rowstride(rs*3//4)
            return True
    elif pixel_format in ("RGBA", "RGBX"):
        checkstride()
        if pixel_format=="RGBA" and "BGRA" in rgb_formats and supports_transparency:
            log("argb_swap: rgba_to_bgra for %s on %s", pixel_format, type(pixels))
            image.set_pixels(rgba_to_bgra(pixels))
            image.set_pixel_format("BGRA")
            image.set_rowstride(rs*3//4)
            return True
        if "RGB" in rgb_formats:
            log("argb_swap: bgrx_to_rgb for %s on %s", pixel_format, type(pixels))
            image.set_pixels(bgrx_to_rgb(pixels))
            image.set_pixel_format("RGB")
            image.set_rowstride(rs*3//4)
            return True
    elif pixel_format=="RGB":
        checkstride(3)
        if "BGRX" in rgb_formats:
            log("argb_swap: rgb_to_bgrx for %s on %s", pixel_format, type(pixels))
            image.set_pixels(rgb_to_bgrx(pixels))
            image.set_pixel_format("BGRX")
            image.set_rowstride(rs//3*4)
            return True
    warning_key = "format-not-handled-%s" % pixel_format
    if first_time(warning_key):
        log.warn("Warning: no matching argb function,")
        log.warn(" cannot convert %s to one of: %s", pixel_format, rgb_formats)
    return False


def bit_to_rectangles(buf, unsigned int w, unsigned int h) -> List[Tuple[int,int,int,int]]:
    cdef const unsigned char* bits
    with buffer_context(buf) as bc:
        bits = <const unsigned char*> (<uintptr_t> int(bc))
        return bitdata_to_rectangles(bits, len(bc), w, h)


cdef object bitdata_to_rectangles(const unsigned char* bitdata, const int bitdata_len,
                           const unsigned int w, const unsigned int h):
    rectangles = []
    cdef unsigned int rowstride = round8up(w)//8
    cdef unsigned char b
    cdef unsigned int start, end, x, y
    for y in range(h):
        x = 0
        b = 0
        while x<w:
            #find the first black pixel,
            if b==0:
                #if there are no left-overs in b then
                #we can move 8 pixels at a time (1 byte):
                while x<w and bitdata[y*rowstride+x//8]==0:
                    x += 8
                if x>=w:
                    break
                #there is a black pixel in this byte (8 pixels):
                b = bitdata[y*rowstride+x//8]
                assert b!=0
            while (b & (1<<(7-x%8)))==0:
                x += 1
            if x>=w:
                break
            start = x
            end = 0
            x += 1
            #find the next white pixel,
            #first, continue searching in the current byte:
            while (x%8)>0:
                if (b & (1<<(7-x%8)))==0:
                    end = x
                    break
                b &= ~(1<<(7-x%8))
                x += 1
            if x>=w:
                end = x = w
            if end==0:
                #now we can move 8 pixels at a time (1 byte):
                while x<w and bitdata[y*rowstride+x//8]==0xff:
                    x += 8
                if x>=w:
                    end = x = w
                else:
                    #there is a white pixel in this byte (8 pixels):
                    b = bitdata[y*rowstride+x//8]
                    while b & (1<<(7-x%8)):
                        #clear this bit so we can continue looking for black pixels in b
                        #when we re-enter the loop at the top
                        b &= ~(1<<(7-x%8))
                        x += 1
                    if x>w:
                        x = w
                    end = x
                    if b==0:
                        x = round8up(x)
            if start<end:
                rectangles.append((start, y, end-start, 1))
    return rectangles


cdef void show_plane_range(name, plane: SizedBuffer, int width, int stride, int height):
    cdef unsigned char minv = 255
    cdef unsigned char maxv = 0
    cdef unsigned char value
    cdef const unsigned char *plane_buf
    with buffer_context(plane) as buf:
        plane_buf = <unsigned char *> (<uintptr_t> int(buf))
        for y in range(height):
            for x in range(width):
                value = plane_buf[y*stride + x]
                if value < minv:
                    minv = value
                if value > maxv:
                    maxv = value
    log.info("%s plane: min=%s, max=%s", name, minv, maxv)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.util.objects import AdHocStruct
from unit.test_util import silence_info
from unit.process_test_util import DisplayContext
from unit.client.mixins.clientmixintest_util import ClientMixinTest


class DisplayClientTest(ClientMixinTest):

    def test_display(self):
        with DisplayContext():
            from xpra.client.mixins import display  # pylint: disable=import-outside-toplevel
            def _DisplayClient():
                dc = display.DisplayClient()
                def get_root_size():
                    return 1024, 768
                dc.get_root_size = get_root_size
                def get_screen_sizes(*_args):
                    return ((1024, 768),)
                dc.get_screen_sizes = get_screen_sizes
                return dc
            opts = AdHocStruct()
            opts.desktop_fullscreen = False
            opts.desktop_scaling = False
            opts.dpi = 144
            opts.refresh_rate = "20"
            with silence_info(display):
                self._test_mixin_class(_DisplayClient, opts, {
                    "display" : ":999",
                    "desktop_size" : (1024, 768),
                    "max_desktop_size" : (3840, 2160),
                    "actual_desktop_size" : (1024, 768),
                    "resize_screen" : True,
                    })

def main():
    unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from collections.abc import Sequence

from xpra.util.str_fn import csv, print_nested_dict
from xpra.os_util import POSIX
from xpra.keyboard.mask import DEFAULT_MODIFIER_MEANINGS, DEFAULT_MODIFIER_NUISANCE
from xpra.scripts.config import TRUE_OPTIONS, FALSE_OPTIONS
from xpra.log import Logger

log = Logger("keyboard")


def get_modifier_names(mod_meanings: dict[str, str]) -> dict[str, str]:
    # modifier names contains the internal modifiers list, ie: "mod1", "control", ...
    # but the user expects the name of the key to be used, ie: "alt" or "super"
    # whereas at best, we keep "Alt_L" : "mod1" mappings... (posix)
    # so generate a map from one to the other:
    modifier_names: dict[str, str] = {}
    meanings = mod_meanings or DEFAULT_MODIFIER_MEANINGS
    DEFAULT_MODIFIER_IGNORE_KEYNAMES = ["Caps_Lock", "Num_Lock", "Scroll_Lock"]
    for pub_name, mod_name in meanings.items():
        if mod_name in DEFAULT_MODIFIER_NUISANCE or pub_name in DEFAULT_MODIFIER_IGNORE_KEYNAMES:
            continue
        # just hope that xxx_L is mapped to the same modifier as xxx_R!
        if pub_name.endswith("_L") or pub_name.endswith("_R"):
            pub_name = pub_name[:-2]
        elif pub_name == "ISO_Level3_Shift":
            pub_name = "AltGr"
        if pub_name not in modifier_names:
            modifier_names[pub_name.lower()] = mod_name
        if pub_name.lower() == "control":
            # alias "control" to "ctrl" as it is often used:
            modifier_names["ctrl"] = mod_name
    log("parse_shortcuts: modifier names=%s", modifier_names)
    return modifier_names


def parse_shortcut_modifiers(s: str, modifier_names: dict[str, str]) -> list[str]:
    # figure out the default shortcut modifiers
    # accept "," or "+" as delimiter:
    shortcut_modifiers = s.lower().replace(",", "+").split("+")

    def mod_defaults():
        mods = ["meta", "shift"]
        if POSIX:
            # gnome intercepts too many of the Alt+Shift shortcuts,
            # so use control with gnome:
            if os.environ.get("XDG_CURRENT_DESKTOP") == "GNOME":
                mods = ["control", "shift"]
        return mods

    if shortcut_modifiers == ["auto"]:
        shortcut_modifiers = mod_defaults()
    elif shortcut_modifiers == ["none"]:
        shortcut_modifiers = []
    else:
        r = []
        for x in shortcut_modifiers:
            x = x.lower()
            if x not in modifier_names:
                log.warn("Warning: invalid shortcut modifier '%s'", x)
            else:
                r.append(x)
        if shortcut_modifiers and not r:
            log.warn(" using defaults instead")
            shortcut_modifiers = mod_defaults()
        else:
            shortcut_modifiers = r
    log("shortcut modifiers=%s", shortcut_modifiers)
    return shortcut_modifiers


def parse_shortcuts(strs: Sequence[str],
                    shortcut_modifiers: Sequence[str],
                    modifier_names: dict[str, str],
                    ) -> dict[str, list[tuple[list[str], str, list[str | float | bool | int | None]]]]:
    """
    if none are defined, add this as default
    it would be nicer to specify it via OptionParser in main,
    but then it would always have to be there with no way of removing it
    whereas now it is enough to define one (any shortcut)
    """
    if not strs:
        strs = ["meta+shift+F4:quit"]
    log("parse_shortcuts(%s)", strs)
    shortcuts: dict[str, list[tuple[list[str], str, list[str | float | bool | int | None]]]] = {}
    # figure out the default shortcut modifiers
    # accept "," or "+" as delimiter:
    for s in strs:
        if s == "none":
            continue
        if s == "clear":
            shortcuts = {}
            continue
        # example for s: Control+F8:some_action()
        if s.find("=") > s.find(":"):
            parts = s.split("=", 1)
        else:
            parts = s.split(":", 1)
        if len(parts) != 2:
            log.error("Error: invalid key shortcut '%s'", s)
            continue
        # example for action: "quit"
        action = parts[1].strip()
        args: list[str | float | bool | int | None] = []
        if action.find("(") > 0 and action.endswith(")"):
            try:
                action, all_args = action[:-1].split("(", 1)
                for x in all_args.split(","):
                    x = x.strip()
                    if not x:
                        continue
                    if (x[0] == '"' and x[-1] == '"') or (x[0] == "'" and x[-1] == "'"):
                        args.append(x[1:-1])
                    elif x == "None":
                        args.append(None)
                    elif x.find(".") != -1:
                        args.append(float(x))
                    elif x.lower() in TRUE_OPTIONS:
                        args.append(True)
                    elif x.lower() in FALSE_OPTIONS:
                        args.append(False)
                    else:
                        args.append(int(x))
            except Exception as e:
                log.warn("Warning: failed to parse arguments of shortcut:")
                log.warn(" '%s': %s", s, e)
                continue
        action = action.replace("-", "_")  # must be an object attribute
        log("action(%s)=%s%s", s, action, args)
        # example for keyspec: ["Alt", "F8"]
        keyspec = parts[0].split("+")
        modifiers = []
        if len(keyspec) > 1:
            valid = True
            # ie: ["Alt"]
            for mod in keyspec[:len(keyspec) - 1]:
                mod = mod.lower()
                if mod == "none":
                    continue
                if mod == "#":
                    # this is the placeholder for the list of shortcut modifiers:
                    for x in shortcut_modifiers:
                        imod = modifier_names.get(x)
                        if imod:
                            modifiers.append(imod)
                    continue
                # find the real modifier for this name:
                # ie: "alt_l" -> "mod1"
                imod = modifier_names.get(mod)
                if not imod:
                    log.warn("Warning: invalid modifier '%s' in keyboard shortcut '%s'", mod, s)
                    log.warn(" the modifiers must be one of: %s", csv(modifier_names.keys()))
                    valid = False
                    break
                modifiers.append(imod)
            if not valid:
                continue
        # should we be validating the keyname?
        keyname = keyspec[len(keyspec) - 1]
        key_shortcuts: list[tuple[list[str], str, list[str | float | bool | int | None]]] = shortcuts.get(keyname, [])
        # remove any existing action using the same modifiers:
        key_shortcuts = [x for x in key_shortcuts if x[0] != modifiers]
        if action != "_":
            key_shortcuts.append((modifiers, action, args))
        shortcuts[keyname] = key_shortcuts
        log("shortcut(%s)=%s", s, csv((modifiers, action, args)))
    log("parse_shortcuts(%s)=%s", strs, shortcuts)
    print_nested_dict(shortcuts, print_fn=log)
    return shortcuts



# This file is part of Xpra.
# Copyright (C) 2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os

from datetime import datetime
from typing import Any
from collections.abc import Sequence, Callable

from xpra.os_util import gi_import
from xpra.scripts.config import make_defaults_struct
from xpra.util.env import osexpand
from xpra.util.parsing import parse_simple_dict
from xpra.util.thread import start_thread
from xpra.log import Logger

log = Logger("util")

OLD_CONFIGURE_TOOL_CONFIG = "99_configure_tool.conf"
CONFIGURE_TOOL_CONFIG = "90_configure_tool.conf"


def get_system_conf_file(dirname="conf.d", filename=CONFIGURE_TOOL_CONFIG) -> str:
    from xpra.platform.paths import get_system_conf_dirs
    return osexpand(os.path.join(get_system_conf_dirs()[0], dirname, filename))


def get_user_config_file(dirname="conf.d", filename=CONFIGURE_TOOL_CONFIG) -> str:
    from xpra.platform.paths import get_user_conf_dirs
    return osexpand(os.path.join(get_user_conf_dirs()[0], dirname, filename))


def parse_user_config_file(dirname="conf.d", filename=CONFIGURE_TOOL_CONFIG) -> dict[str, str | list[str] | dict[str, str]]:
    filename = get_user_config_file(dirname, filename)
    if not os.path.exists(filename):
        return {}
    with open(filename, "r", encoding="utf8") as f:
        data = f.read().replace("\r", "\n")
        return parse_simple_dict(data, sep="\n")


def save_user_config_file(options: dict,
                          dirname="conf.d", filename=CONFIGURE_TOOL_CONFIG) -> None:
    filename = get_user_config_file(dirname, filename)
    conf_dir = os.path.dirname(filename)
    log(f"save_user_config_file({options}, {dirname!r}, {filename!r}) {conf_dir=!r}")
    if not os.path.exists(conf_dir):
        try:
            os.makedirs(conf_dir, mode=0o755)
        except OSError as e:
            log(f"os.makedirs({conf_dir!r}, 0o755)", exc_info=True)
            log.error(f"Error creating configuration directory {conf_dir!r}:")
            log.estr(e)
            return
    with open(filename, "w", encoding="utf8") as f:
        f.write("# generated on " + datetime.now().strftime("%c")+"\n\n")
        for k, v in options.items():
            if isinstance(v, dict):
                for dk, dv in v.items():
                    f.write(f"{k} = {dk}={dv}\n")
                continue
            if not isinstance(v, (list, tuple)):
                v = [v]
            for item in v:
                f.write(f"{k} = {item}\n")


def update_config_attribute(attribute: str, value: str | int | float | list,
                            dirname="conf.d", filename=CONFIGURE_TOOL_CONFIG) -> None:
    update_config_attributes({attribute: value}, dirname, filename)


def update_config_attributes(attributes: dict[str, str | int | float | list],
                             dirname="conf.d", filename=CONFIGURE_TOOL_CONFIG) -> None:
    config = parse_user_config_file(dirname, filename)
    for attribute, value in attributes.items():
        value_str = str(value)
        if isinstance(value, bool):
            value_str = "yes" if bool(value) else "no"
        config[attribute] = value_str
        log(f"update config: {attribute}={value_str}")
    save_user_config_file(config, dirname, filename)


def unset_config_attribute(attribute: str, dirname="conf.d", filename=CONFIGURE_TOOL_CONFIG) -> None:
    unset_config_attributes((attribute, ), dirname, filename)


def unset_config_attributes(attributes: Sequence[str], dirname="conf.d", filename=CONFIGURE_TOOL_CONFIG) -> None:
    config = parse_user_config_file(dirname, filename)
    modified = False
    for attribute in attributes:
        modified |= config.pop(attribute, None) is not None
    if modified:
        save_user_config_file(config, dirname, filename)


def update_config_env(attribute: str, value,
                      dirname="conf.d", filename=CONFIGURE_TOOL_CONFIG) -> None:
    # there can be many env attributes
    log(f"update config env: {attribute}={value}")
    config = parse_user_config_file(dirname, filename)
    env = config.get("env")
    if not isinstance(env, dict):
        log.warn(f"Warning: env option was using invalid type {type(env)}")
        config["env"] = env = {}
    env[attribute] = str(value)
    save_user_config_file(config, dirname, filename)


def get_config_env(var_name: str,
                   dirname="conf.d", filename=CONFIGURE_TOOL_CONFIG) -> str:
    config = parse_user_config_file(dirname, filename)
    env = config.get("env", {})
    if not isinstance(env, dict):
        log.warn(f"Warning: env option was using invalid type {type(env)}")
        return ""
    return env.get(var_name, "")


def with_config(cb: Callable[[Any], bool]) -> None:
    # load config in a thread as this involves IO,
    # then run the callback in the UI thread
    def load_config() -> None:
        defaults = make_defaults_struct()
        GLib = gi_import("GLib")
        GLib.idle_add(cb, defaults)

    start_thread(load_config, "load-config", daemon=True)


def may_migrate() -> None:
    from xpra.platform.paths import get_user_conf_dirs
    for user_conf_dir in get_user_conf_dirs():
        old_config = osexpand(os.path.join(user_conf_dir, "conf.d", OLD_CONFIGURE_TOOL_CONFIG))
        if not os.path.exists(old_config):
            continue
        new_config = osexpand(os.path.join(user_conf_dir, "conf.d", CONFIGURE_TOOL_CONFIG))
        if os.path.exists(new_config):
            continue
        try:
            os.rename(old_config, new_config)
        except OSError:
            pass


may_migrate()



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os.path
import sys
from time import monotonic
from typing import Any
from collections.abc import Sequence

from xpra.util.objects import typedict
from xpra.util.str_fn import csv, Ellipsizer, repr_ellipsized, sorted_nicely, bytestostr, hexstr
from xpra.util.env import envint, first_time
from xpra.common import ConnectionMessage, disconnect_is_an_error, noop
from xpra.os_util import gi_import, get_hex_uuid, POSIX, OSX
from xpra.util.io import stderr_print, load_binary_file
from xpra.net.common import PacketType
from xpra.util.stats import std_unit
from xpra.client.base.client import EXTRA_TIMEOUT
from xpra.client.base.gobject import GObjectXpraClient
from xpra.exit_codes import ExitCode, ExitValue
from xpra.log import Logger

log = Logger("gobject", "client")

GLib = gi_import("GLib")


def errwrite(msg) -> None:
    stderr_print(msg)


class CommandConnectClient(GObjectXpraClient):
    """
        Utility superclass for clients that only send one command
        via the hello packet.
    """

    def __init__(self, opts):
        super().__init__()
        super().init(opts)
        self.display_desc = {}
        # not used by command line clients,
        # so don't try probing for printers, etc
        self.file_transfer = False
        self.printing = False
        self.command_timeout = None
        # don't bother with many of these things for one-off commands:
        for x in ("ui_client", "windows", "webcam", "keyboard", "mouse", "network-state"):
            self.hello_extra[x] = False
        # for newer versions, it is easier:
        self.hello_extra["wants"] = []

    def make_protocol(self, conn):
        self.add_packet_handler("setting-change", noop)
        protocol = super().make_protocol(conn)
        if conn.timeout > 0:
            self.command_timeout = GLib.timeout_add((conn.timeout + self.COMMAND_TIMEOUT) * 1000, self.timeout)
        return protocol

    def timeout(self, *_args) -> None:
        log.warn("timeout!")  # pragma: no cover

    def cancel_command_timeout(self) -> None:
        ct = self.command_timeout
        if ct:
            self.command_timeout = None
            GLib.source_remove(ct)

    def _process_connection_lost(self, packet: PacketType) -> None:
        log("_process_connection_lost%s", packet)
        # override so we don't log a warning
        # "command clients" are meant to exit quickly by losing the connection
        p = self._protocol
        if p and p.input_packetcount == 0:
            # never got any data back, so we have never connected,
            # try to give feedback to the user as to why that is:
            details = ""
            if len(packet) > 1:
                details = ": %s" % csv(packet[1:])
            log.warn("Connection failed%s", details)
            self.quit(ExitCode.CONNECTION_FAILED)
        else:
            self.quit(ExitCode.OK)

    def server_connection_established(self, caps: typedict) -> bool:
        # don't bother parsing the network caps:
        # * it could cause errors if the caps are missing
        # * we don't care about sending anything back after hello
        log("server_capabilities: %s", Ellipsizer(caps))
        log("protocol state: %s", self._protocol.save_state())
        self.init_authenticated_packet_handlers()
        self.cancel_command_timeout()
        self.do_command(caps)
        return True

    def do_command(self, caps: typedict) -> None:
        raise NotImplementedError(f"no do_command method in {self!r}")


class SendCommandConnectClient(CommandConnectClient):
    """
        Utility superclass for clients that only send at least one more packet
        after the hello packet.
        So unlike CommandConnectClient, we do need the network and encryption to be setup.
    """

    def server_connection_established(self, caps) -> bool:
        assert self.parse_encryption_capabilities(caps), "encryption failure"
        assert self.parse_network_capabilities(caps), "network capabilities failure"
        return super().server_connection_established(caps)

    def do_command(self, caps: typedict) -> None:
        raise NotImplementedError()


class HelloRequestClient(SendCommandConnectClient):
    """
        Utility superclass for clients that send a server request
        as part of the hello packet.
    """

    def make_hello_base(self) -> dict[str, Any]:
        return super().make_hello_base() | self.hello_request()

    def timeout(self, *_args) -> None:
        self.warn_and_quit(ExitCode.TIMEOUT, "timeout: server did not disconnect us")

    def hello_request(self):  # pragma: no cover
        raise NotImplementedError(f"`hello_request` not implemented in {self!r}")

    def do_command(self, caps: typedict) -> None:
        self.quit(ExitCode.OK)

    def _process_disconnect(self, packet: PacketType) -> None:
        # overridden method, so we can avoid printing a warning,
        # we haven't received the hello back from the server
        # but that's fine for a request client
        info = tuple(str(x) for x in packet[1:])
        reason = info[0]
        if disconnect_is_an_error(reason):
            self.server_disconnect_warning(*info)
        elif self.exit_code is None:
            # we're not in the process of exiting already,
            # tell the user why the server is disconnecting us
            self.server_disconnect(*info)


class ScreenshotXpraClient(HelloRequestClient):
    """ This client does one thing only:
        it sends the hello packet with a screenshot request
        and exits when the resulting image is received (or timedout)
    """

    def __init__(self, opts, screenshot_filename):
        self.screenshot_filename = screenshot_filename
        super().__init__(opts)

    def hello_request(self) -> dict[str, Any]:
        return {
            "request": "screenshot",
        }

    def do_command(self, caps: typedict) -> None:
        self.send("screenshot")

    def timeout(self, *_args) -> None:
        self.warn_and_quit(ExitCode.TIMEOUT, "timeout: did not receive the screenshot")

    def _process_screenshot(self, packet: PacketType) -> None:
        (w, h, encoding, _, img_data) = packet[1:6]
        assert encoding == "png", "expected png screenshot data but got %s" % encoding
        if not img_data:
            self.warn_and_quit(ExitCode.OK,
                               "screenshot is empty and has not been saved"
                               " (maybe there are no windows or they are not currently shown)")
            return
        if self.screenshot_filename == "-":
            output = os.fdopen(sys.stdout.fileno(), "wb", closefd=False)
        else:
            output = open(self.screenshot_filename, "wb")
        with output:
            output.write(img_data)
            output.flush()
        self.warn_and_quit(ExitCode.OK, f"screenshot {w}x{h} saved to: {self.screenshot_filename!r}")

    def init_packet_handlers(self) -> None:
        super().init_packet_handlers()
        self.add_packets("screenshot")


class InfoXpraClient(CommandConnectClient):
    """ This client does one thing only:
        it queries the server with an 'info' request
    """

    def __init__(self, opts):
        super().__init__(opts)
        self.hello_extra["request"] = "info"

    def timeout(self, *_args):
        self.warn_and_quit(ExitCode.TIMEOUT, "timeout: did not receive the info")

    def do_command(self, caps: typedict) -> None:
        if not caps:
            self.quit(ExitCode.NO_DATA)
            return

        def print_fn(s) -> None:
            sys.stdout.write(f"{s}\n")

        def prettify(k, v):
            ks = str(k)
            if ks.endswith("xid"):
                try:
                    return hex(int(v))
                except (TypeError, ValueError):
                    pass
            if ks.endswith("version") and isinstance(v, (tuple, list)):
                return ".".join(str(x) for x in v)
            if isinstance(v, bytes):
                if ks.endswith(".data"):
                    return hexstr(v)
                return bytestostr(v)
            elif isinstance(v, (tuple, list)):
                eltypes = set(type(x) for x in v)
                if len(eltypes) == 1 and tuple(eltypes)[0] in (int, bool, float):
                    return str(v)
                return type(v)(prettify(k, x) for x in v)
            elif isinstance(v, str) and v.isprintable():
                return v
            return repr(v)

        def print_dict(d: dict, path=""):
            for k in sorted_nicely(d.keys()):
                v = d.get(k)
                fk = prettify(k, k)
                kpath = f"{path}.{fk}".strip(".")
                if isinstance(v, dict):
                    print_dict(v, kpath)
                    continue
                fv = prettify(k, v)
                print_fn(f"{kpath}={fv}")

        exit_code = ExitCode.OK
        try:
            print_dict(dict(caps))
        except OSError:
            exit_code = ExitCode.IO_ERROR
        self.quit(exit_code)


class IDXpraClient(InfoXpraClient):

    def __init__(self, *args):
        super().__init__(*args)
        self.hello_extra["request"] = "id"


class RequestXpraClient(CommandConnectClient):

    def __init__(self, request, opts):
        super().__init__(opts)
        self.hello_extra["request"] = request

    def do_command(self, caps: typedict) -> None:
        self.quit(ExitCode.OK)


class ConnectTestXpraClient(CommandConnectClient):
    """ This client does one thing only:
        it queries the server with an 'info' request
    """

    def __init__(self, opts, **kwargs):
        super().__init__(opts)
        self.value = get_hex_uuid()
        self.hello_extra |= {
            "connect_test_request": self.value,
            "request": "connect_test",
            # tells proxy servers we don't want to connect to the real / new instance:
            "connect": False,
        }
        self.hello_extra.update(kwargs)

    def timeout(self, *_args) -> None:
        self.warn_and_quit(ExitCode.TIMEOUT, "timeout: no server response")

    def _process_connection_lost(self, _packet: PacketType) -> None:
        # we should always receive a hello back and call do_command,
        # which sets the correct exit code, landing here is an error:
        self.quit(ExitCode.FAILURE)

    def do_command(self, caps: typedict) -> None:
        if caps:
            ctr = caps.strget("connect_test_response")
            log("do_command(..) expected connect test response='%s', got '%s'", self.value, ctr)
            if ctr == self.value:
                self.quit(ExitCode.OK)
            else:
                self.quit(ExitCode.INTERNAL_ERROR)
        else:
            self.quit(ExitCode.FAILURE)


class MonitorXpraClient(SendCommandConnectClient):
    """ This client does one thing only:
        it prints out events received from the server.
        If the server does not support this feature it exits with an error.
    """

    def __init__(self, opts):
        super().__init__(opts)
        for x in ("features", "events", "request"):
            self.hello_extra.setdefault("wants", []).append(x)
        self.hello_extra["request"] = "event"

    def timeout(self, *args) -> None:
        pass
        # self.warn_and_quit(ExitCode.TIMEOUT, "timeout: did not receive the info")

    def do_command(self, caps: typedict) -> None:
        log.info("waiting for server events")

    def _process_server_event(self, packet: PacketType) -> None:
        log.info(": ".join(str(x) for x in packet[1:]))

    def init_packet_handlers(self) -> None:
        super().init_packet_handlers()
        self.add_packets("server-event", "ping")

    def _process_ping(self, packet: PacketType) -> None:
        echotime = packet[1]
        self.send("ping_echo", echotime, 0, 0, 0, -1)


class InfoTimerClient(MonitorXpraClient):
    """
        This client keeps monitoring the server
        and requesting info data
    """
    REFRESH_RATE = envint("XPRA_REFRESH_RATE", 1)

    def __init__(self, *args):
        super().__init__(*args)
        self.info_request_pending = False
        self.server_last_info = typedict()
        self.server_last_info_time = 0
        self.info_timer = 0

    def run(self) -> ExitValue:
        from xpra.gtk.signals import register_os_signals
        register_os_signals(self.signal_handler, "")
        v = super().run()
        self.log("run()=%s" % v)
        self.cleanup()
        return v

    def signal_handler(self, signum: int, *args) -> None:
        self.log(f"exit_code={self.exit_code}")
        self.log(f"signal_handler({signum}, {args})")
        self.quit(128 + signum)
        self.log(f"exit_code={self.exit_code}")

    def log(self, message) -> None:
        # this method is overridden in top client to use a log file
        log(message)

    def err(self, e) -> None:
        log.error(str(e))

    def cleanup(self) -> None:
        self.cancel_info_timer()
        MonitorXpraClient.cleanup(self)

    def do_command(self, caps: typedict) -> None:
        self.send_info_request()
        GLib.timeout_add(self.REFRESH_RATE * 1000, self.send_info_request)

    def send_info_request(self, *categories: str) -> bool:
        self.log("send_info_request%s" % (categories,))
        if not self.info_request_pending:
            self.info_request_pending = True
            window_ids = ()  # no longer used or supported by servers
            self.send("info-request", [self.uuid], window_ids, categories)
        if not self.info_timer:
            self.info_timer = GLib.timeout_add((self.REFRESH_RATE + 2) * 1000, self.info_timeout)
        return True

    def init_packet_handlers(self) -> None:
        MonitorXpraClient.init_packet_handlers(self)
        self.add_packets("info-response")

    def _process_server_event(self, packet: PacketType) -> None:
        self.log("server event: %s" % (packet,))
        self.last_server_event = packet[1:]
        self.update_screen()

    def _process_info_response(self, packet: PacketType) -> None:
        self.log("info response: %s" % repr_ellipsized(packet))
        self.cancel_info_timer()
        self.info_request_pending = False
        self.server_last_info = typedict(packet[1])
        self.server_last_info_time = monotonic()
        # log.info("server_last_info=%s", self.server_last_info)
        self.update_screen()

    def cancel_info_timer(self) -> None:
        it = self.info_timer
        if it:
            self.info_timer = 0
            GLib.source_remove(it)

    def info_timeout(self) -> bool:
        self.log("info timeout")
        self.update_screen()
        return True

    def update_screen(self):
        raise NotImplementedError()


def print_prompt() -> None:
    sys.stdout.write("> ")
    sys.stdout.flush()


class ShellXpraClient(SendCommandConnectClient):
    """
        Provides an interactive shell with the socket it connects to
    """

    def __init__(self, opts):
        super().__init__(opts)
        self.stdin_io_watch = None
        self.stdin_buffer = ""
        self.hello_extra["shell"] = "True"

    def timeout(self, *args) -> None:
        """
        The shell client never times out,
        but the superclass calls this method automatically,
        just ignore it.
        """

    def cleanup(self) -> None:
        siw = self.stdin_io_watch
        if siw:
            self.stdin_io_watch = None
            GLib.source_remove(siw)
        super().cleanup()

    def do_command(self, caps: typedict) -> None:
        if not caps.boolget("shell"):
            msg = "this server does not support the 'shell' subcommand"
            log.error(msg)
            self.disconnect_and_quit(ExitCode.UNSUPPORTED, msg)
            return
        # start reading from stdin:
        self.install_signal_handlers()
        stdin = sys.stdin
        fileno = stdin.fileno()
        import fcntl
        fl = fcntl.fcntl(fileno, fcntl.F_GETFL)
        fcntl.fcntl(fileno, fcntl.F_SETFL, fl | os.O_NONBLOCK)
        self.stdin_io_watch = GLib.io_add_watch(sys.stdin,
                                                GLib.PRIORITY_DEFAULT, GLib.IO_IN,
                                                self.stdin_ready)
        print_prompt()

    def stdin_ready(self, *_args) -> bool:
        data = sys.stdin.read()
        # log.warn("stdin=%r", data)
        self.stdin_buffer += data
        sent = 0
        if self.stdin_buffer.endswith("\n"):
            for line in self.stdin_buffer.splitlines():
                if line:
                    if line.rstrip("\n\r") in ("quit", "exit"):
                        self.disconnect_and_quit(ExitCode.OK, "user requested %s" % line)
                        self.stdin_io_watch = None
                        return False
                    self.send("shell-exec", line.encode())
                    sent += 1
        self.stdin_buffer = ""
        if not sent:
            print_prompt()
        return True

    def init_packet_handlers(self) -> None:
        super().init_packet_handlers()
        self.add_packets("shell-reply", "ping")

    def _process_ping(self, packet: PacketType) -> None:
        echotime = packet[1]
        self.send("ping_echo", echotime, 0, 0, 0, -1)

    @staticmethod
    def _process_shell_reply(packet: PacketType) -> None:
        fd = int(packet[1])
        message = packet[2]
        if fd == 1:
            stream = sys.stdout
        elif fd == 2:
            stream = sys.stderr
        else:
            raise ValueError(f"invalid file descriptor f{fd}")
        s = str(message)
        if s.endswith("\n"):
            s = s[:-1]
        stream.write("%s" % s)
        stream.flush()
        if fd == 2:
            stream.write("\n")
            print_prompt()


class VersionXpraClient(HelloRequestClient):
    """ This client does one thing only:
        it queries the server for version information and prints it out
    """

    def hello_request(self) -> dict[str, Any]:
        return {
            "request": "version",
            "full-version-request": True,
        }

    def parse_network_capabilities(self, *_args) -> bool:
        # don't bother checking anything - this could generate warnings
        return True

    def do_command(self, caps: typedict) -> None:
        v = caps.strget("version")
        if not v:
            self.warn_and_quit(ExitCode.FAILURE, "server did not provide the version information")
        else:
            sys.stdout.write(f"{v}\n")
            sys.stdout.flush()
            self.quit(ExitCode.OK)


class ControlXpraClient(CommandConnectClient):
    """ Allows us to send commands to a server.
    """

    def set_command_args(self, command: Sequence[str]) -> None:
        self.command = command

    def timeout(self, *_args) -> None:
        self.warn_and_quit(ExitCode.TIMEOUT, "timeout: server did not respond")

    def do_command(self, caps: typedict) -> None:
        cr = caps.tupleget("command_response")
        if not cr or len(cr) < 2:
            self.warn_and_quit(ExitCode.UNSUPPORTED, "server does not support control command")
            return
        code = int(cr[0])
        text = str(cr[1])
        if code != 0:
            log.warn("server returned error code %s", code)
            self.warn_and_quit(ExitCode.REMOTE_ERROR, " %s" % text)
            return
        self.warn_and_quit(ExitCode.OK, text)

    def make_hello(self) -> dict[str, Any]:
        capabilities = super().make_hello()
        log("make_hello() adding command request '%s' to %s", self.command, capabilities)
        capabilities["command_request"] = tuple(self.command)
        capabilities["request"] = "command"
        return capabilities


class PrintClient(SendCommandConnectClient):
    """
    Allows us to send a file to the server for printing.
    """

    def set_command_args(self, command: Sequence[str]) -> None:
        log("set_command_args(%s)", command)
        self.filename = command[0]
        # print command arguments:
        # filename, file_data, mimetype, source_uuid, title, printer, no_copies, print_options_str = packet[1:9]
        self.command = command[1:]

        if self.filename == "-":
            # replace with filename proposed
            self.filename = command[2]
            # read file from stdin
            with open(sys.stdin.fileno(), mode="rb", closefd=False) as stdin_binary:
                self.file_data = stdin_binary.read()
            log("read %i bytes from stdin", len(self.file_data))
        else:
            if not self.validate_file_size(os.path.getsize(self.filename)):
                return
            self.file_data = load_binary_file(self.filename)
            log("read %i bytes from %s", len(self.file_data), self.filename)
        if not self.validate_file_size(len(self.file_data)):
            return
        assert self.file_data, "no data found for '%s'" % self.filename

    def validate_file_size(self, size: int) -> bool:
        if size <= self.file_size_limit:
            return True
        self.warn_and_quit(ExitCode.FILE_TOO_BIG,
                           "the file is too large: %sB (the file size limit is %sB)" % (
                               std_unit(size), std_unit(self.file_size_limit)))
        return False

    def client_type(self) -> str:
        return "Python/GObject/Print"

    def timeout(self, *_args) -> None:
        self.warn_and_quit(ExitCode.TIMEOUT, "timeout: server did not respond")

    def do_command(self, caps: typedict) -> None:
        printing = caps.boolget("printing")
        if not printing:
            self.warn_and_quit(ExitCode.UNSUPPORTED, "server does not support printing")
            return
        # we don't compress file data
        # (this should run locally most of the time anyway)
        from xpra.net.compression import Compressed  # pylint: disable=import-outside-toplevel
        blob = Compressed("print", self.file_data)
        self.send("print", self.filename, blob, *self.command)
        log("print: sending %s as %s for printing", self.filename, blob)
        GLib.idle_add(self.send, "disconnect", ConnectionMessage.DONE.value, "detaching")

    def make_hello(self) -> dict[str, Any]:
        capabilities = super().make_hello()
        capabilities.setdefault("wants", []).append("features")
        capabilities["request"] = "print"
        return capabilities


class ExitXpraClient(HelloRequestClient):
    """ This client does one thing only:
        it asks the server to terminate (like stop),
        but without killing the Xvfb or clients.
    """

    def hello_request(self) -> dict[str, Any]:
        return {
            "request": "exit",
        }

    def do_command(self, caps: typedict) -> None:
        GLib.idle_add(self.send,
                      "exit-server",
                      os.environ.get("XPRA_EXIT_MESSAGE", ConnectionMessage.SERVER_EXIT.value)
                      )


class StopXpraClient(HelloRequestClient):
    """ stop a server """

    def hello_request(self) -> dict[str, Any]:
        return {
            "request": "stop",
        }

    def do_command(self, caps: typedict) -> None:
        if not self.server_client_shutdown:
            log.error("Error: cannot shutdown this server")
            log.error(" the feature is disabled on the server")
            self.quit(ExitCode.UNSUPPORTED)
            return
        # with newer (v5) servers, the "stop" request should have done it,
        # but let's try harder for older servers:
        GLib.timeout_add(1000, self.send_shutdown_server)
        # GLib.idle_add(self.send_shutdown_server)
        # not exiting the client here,
        # the server should send us the shutdown disconnection message anyway
        # and if not, we will then hit the timeout to tell us something went wrong


class DetachXpraClient(HelloRequestClient):
    """ run the `detach` subcommand """

    def hello_request(self) -> dict[str, Any]:
        return {
            "request": "detach",
        }

    def do_command(self, caps: typedict) -> None:
        GLib.idle_add(self.send, "disconnect", ConnectionMessage.DONE.value, "detaching")
        # not exiting the client here,
        # the server should disconnect us with the response


class WaitForDisconnectXpraClient(DetachXpraClient):
    """ we just want the connection to close """

    def _process_disconnect(self, _packet: PacketType) -> None:
        self.quit(ExitCode.OK)


class RequestStartClient(HelloRequestClient):
    """ request the system proxy server to start a new session for us """
    # wait longer for this command to return:
    from xpra.scripts.main import WAIT_SERVER_TIMEOUT
    COMMAND_TIMEOUT = EXTRA_TIMEOUT + WAIT_SERVER_TIMEOUT

    def dots(self) -> bool:
        errwrite(".")
        return not self.connection_established

    def _process_connection_lost(self, packet: PacketType) -> None:
        errwrite("\n")
        super()._process_connection_lost(packet)

    def hello_request(self) -> dict[str, Any]:
        if first_time("hello-request"):
            # this can be called again if we receive a challenge,
            # but only print this message once:
            errwrite("requesting new session, please wait")
        GLib.timeout_add(1 * 1000, self.dots)
        return {
            "start-new-session": self.start_new_session,
            # tells proxy servers we don't want to connect to the real / new instance:
            "connect": False,
        }

    def server_connection_established(self, caps: typedict) -> bool:
        # the server should respond with the display chosen
        log("server_connection_established() exit_code=%s", self.exit_code)
        display = caps.strget("display")
        if display:
            mode = caps.strget("mode")
            session_type = {
                "start": "seamless ",
                "start-desktop": "desktop ",
                "shadow": "shadow ",
            }.get(mode, "")
            try:
                errwrite("\n%ssession now available on display %s\n" % (session_type, display))
                if POSIX and not OSX and self.displayfd > 0 and display and display.startswith(":"):
                    from xpra.platform.displayfd import write_displayfd
                    log("writing display %s to displayfd=%s", display, self.displayfd)
                    write_displayfd(self.displayfd, display[1:])
            except OSError:
                log("server_connection_established(..)", exc_info=True)
        if not self.exit_code:
            self.quit(0)
        return True

    def __init__(self, opts):
        super().__init__(opts)
        try:
            self.displayfd = int(opts.displayfd)
        except (ValueError, TypeError):
            self.displayfd = 0



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import unittest

from xpra.util.objects import typedict
from xpra.net.file_transfer import (
    basename, safe_open_download_file,
    FileTransferAttributes, FileTransferHandler,
    get_open_env,
)


class TestVersionUtilModule(unittest.TestCase):

    def test_basename(self):
        def t(s, e):
            r = basename(s)
            assert r==e, "expected '%s' but got '%s' for '%s'" % (r, e, s)
        t("hello", "hello")
        t("/path/to/foo", "foo")
        t("\\other\\path\\bar", "bar")

    def test_safe_open(self):
        filename, fd = safe_open_download_file("hello", "application/pdf")
        try:
            dupe_filename, dupe_fd = safe_open_download_file("hello", "application/pdf")
            assert dupe_filename!=filename
            try:
                os.close(dupe_fd)
            finally:
                os.unlink(dupe_filename)
            os.close(fd)
        finally:
            os.unlink(filename)

    def test_file_transfer_attributes(self):
        fta = FileTransferAttributes()
        assert fta.get_file_transfer_features()
        assert fta.get_info()

    def test_file_transfer_handler(self):
        fth = FileTransferHandler()
        fth.init_attributes()
        assert get_open_env()
        caps = typedict()
        fth.parse_file_transfer_caps(caps)
        assert fth.get_info()
        fth.cleanup()


def main():
    unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from collections.abc import Iterable

from xpra.util.io import is_socket
from xpra.log import Logger

log = Logger("server", "ssh")


def ssh_dir_path(session_dir: str = os.environ.get("XPRA_SESSION_DIR", "")) -> str:
    return os.path.join(session_dir, "ssh")


def setup_ssh_auth_sock(session_dir: str) -> str:
    # the 'ssh' dir contains agent socket symlinks to the real agent socket,
    # so we can just update the "agent" symlink
    # which is the one that applications are told to use
    ssh_dir = ssh_dir_path(session_dir)
    if not os.path.exists(ssh_dir):
        os.mkdir(ssh_dir, 0o700)
    # ie: "/run/user/1000/xpra/10/ssh/agent"
    agent_sockpath = get_ssh_agent_path("agent")
    # the current value from the environment:
    # ie: "SSH_AUTH_SOCK=/tmp/ssh-XXXX4KyFhe/agent.726992"
    # or "SSH_AUTH_SOCK=/run/user/1000/keyring/ssh"
    cur_sockpath = os.environ.pop("SSH_AUTH_SOCK", None)
    # ie: "/run/user/1000/xpra/10/ssh/agent.default"
    agent_default_sockpath = get_ssh_agent_path("agent.default")
    if os.path.islink(agent_default_sockpath):
        if not os.path.exists(agent_default_sockpath):
            # remove dead symlink
            os.unlink(agent_default_sockpath)
    elif os.path.exists(agent_default_sockpath):
        raise RuntimeError(f"{agent_default_sockpath!r} already exists but it is not a symbolic link")
    if cur_sockpath and cur_sockpath != agent_sockpath and not os.path.exists(agent_default_sockpath):
        # the current agent socket will be the default:
        # ie: "agent.default" -> "/run/user/1000/keyring/ssh"
        os.symlink(cur_sockpath, agent_default_sockpath)
    set_ssh_agent()
    return agent_sockpath


def get_ssh_agent_path(filename: str, session_dir: str = os.environ.get("XPRA_SESSION_DIR", "")) -> str:
    ssh_dir = ssh_dir_path(session_dir)
    if "/" in filename or ".." in filename:
        raise ValueError(f"illegal characters found in ssh agent filename {filename!r}")
    return os.path.join(ssh_dir, filename or "agent.default")


def set_ssh_agent(filename: str = "") -> None:
    ssh_dir = ssh_dir_path()
    if filename and os.path.isabs(filename):
        sockpath = filename
    else:
        filename = filename or "agent.default"
        sockpath = get_ssh_agent_path(filename)
    if not os.path.exists(sockpath):
        log(f"set_ssh_agent: invalid sockpath {sockpath!r}")
        return
    agent_sockpath = os.path.join(ssh_dir, "agent")
    try:
        if os.path.islink(agent_sockpath):
            os.unlink(agent_sockpath)
        log("setting ssh agent link:")
        log(f" {filename!r} -> {agent_sockpath!r}")
        os.symlink(filename, agent_sockpath)
    except OSError as e:
        log(f"set_ssh_agent({filename})", exc_info=True)
        log.error(f"Error: failed to set ssh agent socket path to {filename!r}")
        log.estr(e)


def clean_agent_socket(uuid: str = "") -> None:
    sockpath = get_ssh_agent_path(uuid)
    try:
        if os.path.exists(sockpath):
            log(f"removing ssh agent socket {sockpath!r}")
            os.unlink(sockpath)
    except OSError as e:
        log.error(f"Error: failed to remove ssh agent socket path {sockpath!r}")
        log.error(f" for uuid {uuid!r}")
        log.estr(e)


def setup_proxy_ssh_socket(
        cmdline: Iterable[str],
        auth_sock: str = os.environ.get("SSH_AUTH_SOCK", ""),
        session_dir: str = os.environ.get("XPRA_SESSION_DIR", ""),
) -> str:
    log(f"setup_proxy_ssh_socket({cmdline}, {auth_sock!r}")
    # this is the socket path that the ssh client wants us to use:
    # ie: "SSH_AUTH_SOCK=/tmp/ssh-XXXX4KyFhe/agent.726992"
    if not auth_sock or not os.path.exists(auth_sock) or not is_socket(auth_sock):
        log(f"setup_proxy_ssh_socket invalid SSH_AUTH_SOCK={auth_sock!r}")
        return ""
    if not session_dir or not os.path.exists(session_dir) or not os.path.isdir(session_dir):
        log(f"setup_proxy_ssh_socket invalid session_dir={session_dir!r}")
        return ""
    # locate the ssh agent uuid,
    # which is used to derive the agent path symlink
    # that the server will want to use for this connection,
    # newer clients pass it to the remote proxy command process using an env var:
    agent_uuid = None
    for x in cmdline:
        if x.startswith("--env=SSH_AGENT_UUID="):
            agent_uuid = x[len("--env=SSH_AGENT_UUID="):]
            break
    # prevent illegal paths:
    if not agent_uuid or agent_uuid.find("/") >= 0 or agent_uuid.find(".") >= 0:
        log(f"setup_proxy_ssh_socket invalid SSH_AGENT_UUID={agent_uuid!r}")
        return ""
    # ie: "/run/user/$UID/xpra/$DISPLAY/ssh/$UUID
    agent_uuid_sockpath = get_ssh_agent_path(agent_uuid, session_dir)
    if os.path.exists(agent_uuid_sockpath) or os.path.islink(agent_uuid_sockpath):
        if is_socket(agent_uuid_sockpath):
            log(f"setup_proxy_ssh_socket keeping existing valid socket {agent_uuid_sockpath!r}")
            # keep the existing socket unchanged - somehow it still works?
            return agent_uuid_sockpath
        log(f"setup_proxy_ssh_socket removing invalid symlink / socket {agent_uuid_sockpath!r}")
        try:
            os.unlink(agent_uuid_sockpath)
        except OSError as e:
            log(f"os.unlink({agent_uuid_sockpath!r})", exc_info=True)
            log.error("Error: removing the broken ssh agent symlink")
            log.estr(e)
    log(f"setup_proxy_ssh_socket {agent_uuid_sockpath!r} -> {auth_sock!r}")
    try:
        os.symlink(auth_sock, agent_uuid_sockpath)
    except OSError as e:
        log(f"os.symlink({auth_sock}, {agent_uuid_sockpath})", exc_info=True)
        log.error("Error creating ssh agent socket symlink")
        log.estr(e)
        return ""
    return agent_uuid_sockpath


def setup_client_ssh_agent_socket(uuid: str, ssh_auth_sock: str) -> str:
    if not uuid:
        log("cannot setup ssh agent without client uuid")
        return ""
    # perhaps the agent sock path for this uuid already exists:
    # ie: /run/user/1000/xpra/10/$UUID
    sockpath = get_ssh_agent_path(uuid)
    log(f"get_ssh_agent_path({uuid})={sockpath}")
    if not os.path.exists(sockpath) or not is_socket(sockpath):
        if os.path.islink(sockpath):
            # dead symlink
            try:
                os.unlink(sockpath)
            except OSError as e:
                log(f"os.unlink({sockpath!r})", exc_info=True)
                log.error("Error: removing the broken ssh agent symlink")
                log.estr(e)
        # perhaps this is a local client,
        # and we can find its agent socket and create the symlink now:
        log(f"client supplied ssh-auth-sock={ssh_auth_sock}")
        if ssh_auth_sock and os.path.isabs(ssh_auth_sock) and os.path.exists(ssh_auth_sock) and is_socket(
                ssh_auth_sock):  # noqa: E501
            try:
                # ie: /run/user/1000/xpra/10/$UUID -> /tmp/ssh-XXXXvjt4hN/agent.766599
                os.symlink(ssh_auth_sock, sockpath)
            except OSError as e:
                log(f"os.symlink({ssh_auth_sock}, {sockpath})", exc_info=True)
                log.error("Error setting up ssh agent socket symlink")
                log.estr(e)
                sockpath = ""
    return sockpath



# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections.abc import Callable

from xpra.auth.sys_auth_base import SysAuthenticator


class Authenticator(SysAuthenticator):

    def __init__(self, **kwargs):
        self.value: str = kwargs.pop("value", "")
        super().__init__(**kwargs)
        self.authenticate_check: Callable = self.authenticate_hmac

    def __repr__(self):
        return "password"

    def get_password(self) -> str:
        return self.value



# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.auth.sys_auth_base import SysAuthenticator, log
from xpra.util.objects import typedict


class Authenticator(SysAuthenticator):

    def __init__(self, **kwargs):
        log("capability.Authenticator(%s)", kwargs)
        self.uid = -1
        self.gid = -1
        self.property = kwargs.pop("property", "display")
        self.value = str(kwargs.pop("value", ""))
        # connection = kwargs.get("connection", None)
        super().__init__(**kwargs)

    def get_uid(self) -> int:
        return self.uid

    def get_gid(self) -> int:
        return self.gid

    def requires_challenge(self) -> bool:
        return False

    def authenticate(self, caps: typedict) -> bool:  # pylint: disable=arguments-differ
        value = caps.strget(self.property, "")
        log("capability.authenticate(..) %r=%r (value required: %r)",
            self.property, value, self.value)
        return value == self.value

    def __repr__(self):
        return "capability"



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections.abc import Callable
from ctypes import (
    sizeof, byref,
    WinError, get_last_error,  # @UnresolvedImport
)

from xpra.util.env import envbool
from xpra.platform.win32.wtsapi import (
    NOTIFY_FOR_THIS_SESSION,
    WM_WTSSESSION_CHANGE, WM_DWMNCRENDERINGCHANGED, WM_DWMCOMPOSITIONCHANGED,
    WTSRegisterSessionNotification, WTSUnRegisterSessionNotification,
)
from xpra.platform.win32.wndproc_events import WNDPROC_EVENT_NAMES
from xpra.platform.win32 import constants as win32con
from xpra.platform.win32.common import (
    WNDCLASSEX, WNDPROC,
    RegisterClassExW,
    CreateWindowExA, DestroyWindow,
    UnregisterClassW, DefWindowProcW,
    GetModuleHandleA,
    GetStockObject,
)
from xpra.log import Logger

log = Logger("events", "win32")

KNOWN_EVENTS = {}
POWER_EVENTS = {}
for x in dir(win32con):
    if x.endswith("_EVENT"):
        v = getattr(win32con, x)
        KNOWN_EVENTS[v] = x
    if x.startswith("PBT_"):
        v = getattr(win32con, x)
        POWER_EVENTS[v] = x

IGNORE_EVENTS = {
    win32con.WM_ACTIVATEAPP: "WM_ACTIVATEAPP",
    win32con.WM_TIMECHANGE: "WM_TIMECHANGE",
    win32con.WM_DESTROY: "WM_DESTROY",
    win32con.WM_COMMAND: "WM_COMMAND",
    win32con.WM_DEVICECHANGE: "WM_DEVICECHANGE",
    win32con.WM_DISPLAYCHANGE: "WM_DISPLAYCHANGE",  # already taken care of by gtk event
    win32con.WM_NCCALCSIZE: "WM_NCCALCSIZE",  # happens after resume too?
    win32con.WM_WINDOWPOSCHANGED: "WM_WINDOWPOSCHANGED",  # happens after resume too?
    win32con.WM_WININICHANGE: "WM_WININICHANGE",  # happens after resume too?
    win32con.WM_WINDOWPOSCHANGING: "WM_WINDOWPOSCHANGING",
    win32con.WM_GETMINMAXINFO: "WM_GETMINMAXINFO",
    win32con.WM_SYSCOLORCHANGE: "WM_SYSCOLORCHANGE",
    WM_WTSSESSION_CHANGE: "WM_WTSSESSION_CHANGE",
    WM_DWMNCRENDERINGCHANGED: "WM_DWMNCRENDERINGCHANGED",
    800: "screen background changed",  # I can't find this definition anywhere
    win32con.WM_SIZE: "WM_SIZE: screen resized",  # we get a GTK signal for this
}
LOG_EVENTS = {
    win32con.WM_POWERBROADCAST: "WM_POWERBROADCAST: power management event",
    win32con.WM_TIMECHANGE: "WM_TIMECHANGE: time change event",
    win32con.WM_INPUTLANGCHANGE: "WM_INPUTLANGCHANGE: input language changed",
    WM_DWMCOMPOSITIONCHANGED: "WM_DWMCOMPOSITIONCHANGED: Desktop Window Manager composition has been enabled or disabled",
}
KNOWN_WM_EVENTS = IGNORE_EVENTS.copy()
KNOWN_WM_EVENTS.update(WNDPROC_EVENT_NAMES)
NIN_BALLOONSHOW = win32con.WM_USER + 2
NIN_BALLOONHIDE = win32con.WM_USER + 3
NIN_BALLOONTIMEOUT = win32con.WM_USER + 4
NIN_BALLOONUSERCLICK = win32con.WM_USER + 5
BALLOON_EVENTS = {
    NIN_BALLOONSHOW: "NIN_BALLOONSHOW",
    NIN_BALLOONHIDE: "NIN_BALLOONHIDE",
    NIN_BALLOONTIMEOUT: "NIN_BALLOONTIMEOUT",
    NIN_BALLOONUSERCLICK: "NIN_BALLOONUSERCLICK",
}
KNOWN_WM_EVENTS.update(BALLOON_EVENTS)

# anything else we don't have yet:
for x in dir(win32con):
    if x.startswith("WM_") and x not in KNOWN_WM_EVENTS:
        v = getattr(win32con, x)
        KNOWN_WM_EVENTS[v] = x

WINDOW_EVENTS = envbool("XPRA_WIN32_WINDOW_EVENTS", True)

EVENT_CALLBACK_TYPE = Callable[[int, int], None]


def add_handler(event: str, handler: Callable) -> None:
    pass


def remove_handler(event: str, handler: Callable) -> None:
    pass


class Win32Eventlistener:

    def __init__(self):
        assert singleton is None
        self.hwnd = None
        self.event_callbacks: dict[int, list[EVENT_CALLBACK_TYPE]] = {}
        self.ignore_events = IGNORE_EVENTS
        self.log_events = LOG_EVENTS

        if not WINDOW_EVENTS:
            return

        self.wc = WNDCLASSEX()
        self.wc.cbSize = sizeof(WNDCLASSEX)
        self.wc.style = win32con.CS_GLOBALCLASS | win32con.CS_VREDRAW | win32con.CS_HREDRAW
        self.wc.lpfnWndProc = WNDPROC(self.WndProc)
        self.wc.cbClsExtra = 0
        self.wc.cbWndExtra = 0
        self.wc.hInstance = GetModuleHandleA(0)
        self.wc.hIcon = 0
        self.wc.hCursor = 0
        self.wc.hBrush = GetStockObject(win32con.WHITE_BRUSH)
        self.wc.lpszMenuName = 0
        self.wc.lpszClassName = "Xpra-Event-Window"
        self.wc.hIconSm = 0
        self.wc.hbrBackground = win32con.COLOR_WINDOW
        self.wc_atom = RegisterClassExW(byref(self.wc))
        if self.wc_atom == 0:
            raise WinError(get_last_error())

        self.hwnd = CreateWindowExA(0, self.wc_atom, "For xpra event listener only",
                                    win32con.WS_CAPTION,
                                    0, 0, win32con.CW_USEDEFAULT, win32con.CW_USEDEFAULT,
                                    0, 0, self.wc.hInstance, None)
        if self.hwnd == 0:
            raise WinError(get_last_error())

        # register our interest in session events:
        # http://timgolden.me.uk/python/win32_how_do_i/track-session-events.html#isenslogon
        # http://stackoverflow.com/questions/365058/detect-windows-logout-in-python
        # http://msdn.microsoft.com/en-us/library/aa383841.aspx
        # http://msdn.microsoft.com/en-us/library/aa383828.aspx
        WTSRegisterSessionNotification(self.hwnd, NOTIFY_FOR_THIS_SESSION)
        log("Win32Eventlistener created with hwnd=%s", self.hwnd)

    def cleanup(self):
        log("Win32Eventlistener.cleanup()")
        self.event_callbacks = {}

        hwnd = self.hwnd
        if hwnd:
            WTSUnRegisterSessionNotification(hwnd)

            self.hwnd = None
            try:
                DestroyWindow(hwnd)
            except Exception as e:
                log.error("Error during cleanup of event window instance:")
                log.estr(e)

            try:
                UnregisterClassW(self.wc.lpszClassName, self.wc.hInstance)
            except Exception as e:
                log.error("Error during cleanup of event window class:")
                log.estr(e)

    def add_event_callback(self, event: int, callback: EVENT_CALLBACK_TYPE):
        self.event_callbacks.setdefault(event, []).append(callback)

    def remove_event_callback(self, event: int, callback: EVENT_CALLBACK_TYPE):
        callbacks = self.event_callbacks.get(event, [])
        if callback in callbacks:
            callbacks.remove(callback)

    def WndProc(self, hWnd: int, msg: int, wParam: int, lParam: int):
        callbacks = self.event_callbacks.get(msg)
        event_name = KNOWN_WM_EVENTS.get(msg, hex(msg))
        log("callbacks for event %s: %s", event_name, callbacks)
        if hWnd == self.hwnd:
            if callbacks:
                for c in callbacks:
                    with log.trap_error("Error on event callback %s", c):
                        c(wParam, lParam)
            elif msg in self.ignore_events:
                log("%s: %s / %s", self.ignore_events.get(msg), wParam, lParam)
            elif msg in self.log_events:
                log.info("%s: %s / %s", self.log_events.get(msg), wParam, lParam)
            else:
                log_fn = log.warn
                if 0 <= msg <= win32con.WM_USER or msg > 0xFFFF:
                    ut = "reserved system"
                elif win32con.WM_USER <= msg <= 0x7FFF:
                    ut = "WM_USER"
                elif 0x8000 <= msg <= 0xBFFF:
                    ut = "WM_APP"
                elif 0xC000 <= msg <= 0xFFFF:
                    ut = "string"
                    log_fn = log.debug
                else:
                    ut = "/ unexpected"
                log_fn("unknown %s message: %s / %#x / %#x", ut, event_name, int(wParam), int(lParam))
            if msg == win32con.WM_DESTROY:
                self.cleanup()
        elif self.hwnd and hWnd:
            log.warn("invalid hwnd: %s (expected %s)", hWnd, self.hwnd)
        r = DefWindowProcW(hWnd, msg, wParam, lParam)
        log("DefWindowProc%s=%s", (hWnd, msg, wParam, lParam), r)
        return r


singleton: Win32Eventlistener | None = None


def get_win32_event_listener(create=True) -> Win32Eventlistener | None:
    global singleton
    if not singleton and create:
        singleton = Win32Eventlistener()
    return singleton



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
from collections.abc import Sequence

from xpra.auth.sys_auth_base import SysAuthenticatorBase, xor, log
from xpra.auth.common import parse_uid, parse_gid
from xpra.net.digest import get_salt, get_digests, gendigest
from xpra.util.objects import typedict
from xpra.os_util import WIN32
from xpra.util.io import stderr_print


class Authenticator(SysAuthenticatorBase):
    CLIENT_USERNAME = True

    def __init__(self, **kwargs):
        self.service = kwargs.pop("service", "")
        self.uid = parse_uid(kwargs.pop("uid", None))
        self.gid = parse_gid(kwargs.pop("gid", None))
        kwargs["prompt"] = kwargs.pop("prompt", "kerberos token")
        super().__init__(**kwargs)
        log("kerberos-token auth: service=%r, username=%r", self.service, kwargs.get("username"))

    def get_uid(self) -> int:
        return self.uid

    def get_gid(self) -> int:
        return self.gid

    def __repr__(self):
        return "kerberos-token"

    def get_challenge(self, digests: Sequence[str]) -> tuple[bytes, str]:
        assert not self.challenge_sent
        self.req_challenge(digests, "kerberos")
        self.salt = get_salt()
        self.challenge_sent = True
        return self.salt, f"kerberos:{self.service}"

    def check_password(self, token: str) -> bool:
        log("check(%r)", token)
        assert self.challenge_sent
        try:
            if WIN32:
                import winkerberos as kerberos
            else:
                import kerberos  # @Reimport
        except ImportError as e:
            log("check(..)", exc_info=True)
            log.warn("Warning: cannot use kerberos token authentication:")
            log.warn(" %s", e)
            return False
        v, ctx = kerberos.authGSSServerInit(self.service)  # @UndefinedVariable
        if v != 1:
            log.error("Error: kerberos GSS server init failed for service '%s'", self.service)
            return False
        try:
            r = kerberos.authGSSServerStep(ctx, token)  # @UndefinedVariable
            log("kerberos auth server step result: %s", r == 1)
            if r != 1:
                return False
            targetname = kerberos.authGSSServerTargetName(ctx)  # @UndefinedVariable
            # response = kerberos.authGSSServerResponse(ctx)
            principal = kerberos.authGSSServerUserName(ctx)  # @UndefinedVariable
            # ie: user1@LOCALDOMAIN
            # maybe we should validate the realm?
            log("kerberos targetname=%s, principal=%s", targetname, principal)
            return True
        finally:
            kerberos.authGSSServerClean(ctx)  # @UndefinedVariable


def main(argv) -> int:
    # pylint: disable=import-outside-toplevel
    from xpra.platform import program_context
    with program_context("Kerberos-Token-Auth", "Kerberos Token Authentication"):
        if len(argv) != 3:
            stderr_print("%s invalid arguments" % argv[0])
            stderr_print("usage: %s username token" % argv[0])
            return 1
        username = argv[1]
        token = argv[2]
        kwargs = {"username": username}
        a = Authenticator(**kwargs)
        server_salt, digest = a.get_challenge(("xor", ))
        salt_digest = a.choose_salt_digest(get_digests())
        assert digest == "xor"
        client_salt = get_salt(len(server_salt))
        combined_salt = gendigest(salt_digest, client_salt, server_salt)
        response = xor(token, combined_salt)
        caps = typedict({
            "challenge_response": response,
            "challenge_client_salt": client_salt,
        })
        a.authenticate(caps)
    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv))
